#!/usr/bin/python

####
# Re-implementation of the original SALT2 model.
####

import logging

import numpy as np
import scipy.sparse
import sncosmo

try:
    from sksparse.cholmod import cholesky_AAt
except ImportError:
    from scikits.sparse.cholmod import cholesky_AAt

from nacl.instruments import FilterDb
from nacl.lib import bspline
from nacl.lib.fitparameters import FitParameters
from nacl.lib.instruments import MagSys
from . import colorlaw, lightcurves, spectra

logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                    level=logging.INFO)

from lemaitre import bandpasses
from nacl import bandpasses as bp

class SALT2Like(object):
    """A re-implementation of the SALT2 model

    SALT2 is an empirical SN spectrophotometric model. This class provides a
    pure python re-implementation of SALT2 with a number of improvements.

    The SALT2 parametrization is defined as follows. In the SN restframe the
    absolute spectral density :math:`S(\lambda)` is:

    .. math::

           S(\lambda, \mathrm{p}) = X_0 \times \left[M_0(\lambda, \mathrm{p}) + X_1
           \times M_1(\lambda, \mathrm{p})\right]\ 10^{0.4 c CL(\lambda)}

    where :math:`\mathrm{p}` is the SN phase, i.e. the restframe time since SN
    peak luminosity:

        .. math::

            \mathrm{ph} = \frac{t_{MJD} - t_{max}}{1 + z}

    :math:`M_0`, :math:`M_1` are global surfaces describing the spectral
    evolution of the average SN and its principal variation, and
    :math:`CL(\lambda)` is a global "extinction correction" describing the
    color diversity of the SNIa family.

    :math:`(X_0, X_1, c)` are SN-dependent. :math:`X_0` is the amplitude of the
    "SN-frame B-band lightcurve" as inferred from the observer-frame fluxes.
    :math:`X_1` is the coordinate of the SN along :math:`M_1`. In practice, it
    quantifies the variations of the light curve width. :math:`c` is the SN
    color -- or more exactly, the SN color excess with respect to the average
    SN color.

    Each Light Curve are evaluated interactively to use multi-threading and
    spectra are evaluated simultaneously.
    """
    def __init__(self, training_dataset,
                 wl_range=(2000., 9000.),  # . 11000.),
                 phase_range=(-20., 50.), 
                 basis_knots=[127, 20],
                 basis_filter_knots=900,
                 wl_grid=None, phase_grid=None,
                 spectrum_recal_degree=3,
                 init_from_salt2_file=None,
                 init_from_training_dataset=False,
                 normalization_band_name='SWOPE::B'):
        """Constructor.
              - instantiate the bases (phase, wavelength, filter)
              - compute the Grams
              - compute the lambda_eff matrix
              - instantiate a color law and evaluate it on the lambda_eff's

        Parameters
        ----------
        training_dataset : nacl.dataset.TrainingDataset
          training dataset.
        phase_range : 2-tuple
          nominal phase range of the model.
        wl_range : 2-tuple
          nominal wavelength range of the model.
        basis_knots : 2-list
          number of wavelength and photometric knot.
        basis_filter_knots : int
          number of knot for the filter basis
        spectrum_recal_degree : int
          degree of the spectrum recalibration polynomials
        init_from_salt2_file : str
          uf the model is to be initialized with the original
          SALT2-4 surfaces and bases, get them from this file.
        normalization_band_name : str
          use this band to normalize the model.
        """
        self.training_dataset = training_dataset
        self.lc_data = training_dataset.lc_data
        self.phase_range = phase_range
        self.wl_range = wl_range
        self.n_wl, self.n_ph = basis_knots[0],  basis_knots[1]
        self.basis_knots = basis_knots
        self.basis_filter_knots = basis_filter_knots
        self.spectrum_recal_degree = spectrum_recal_degree
        self.delta_phase = 0. # check: do not remember what this is
        self.filterpath = self.training_dataset.filterpath
        if self.training_dataset.lc_data is not None:
            self.bands = list(self.training_dataset.transmissions.keys())
        self.normalization_band_name = normalization_band_name
        # self.calib_variance = calib_variance

        self.timing = []
        
        self.y = training_dataset.get_all_fluxes()
        self.yerr = training_dataset.get_all_fluxerr()
        self.bads = training_dataset.get_valid()==0

        self.basis, self.gram, self.G2, self.L_eff = None, None, None, None
        self.val, self.jacobian_val, self.jacobian_i, self.jacobian_j = None, None, None, None
        self.polynome_color_law, self.jacobian_color_law = None, None

        # Filter database
        # TODO: we should be able to optimize this a little
        # (i.e. adapt the filter bases, to lower the size of the grams)
        self.filter_basis = bspline.BSpline(np.linspace(self.wl_range[0], self.wl_range[1],  # 2000., 9000.,  #
                                                        basis_filter_knots), order=4)  # was 9000
        try:
            self.filter_db = FilterDb(self.filter_basis, self.training_dataset,
                                      additional_band=[normalization_band_name],
                                      filterpath=self.filterpath)
        except:
            self.filter_db = bp.FilterDb(self.filter_basis, self.training_dataset, bands=['standard::b'])
        self.color_law = colorlaw.ColorLaw()

        n_lc_meas, n_spec_meas, n_spectrophot_meas = \
            self.training_dataset.nb_meas(split_by_type=1)

        # lc = self.training_dataset.lc_data
        # d = dict({(lc[i]['band_id'], lc[i]['Filter']) for i in np.arange(len(lc))})
        # self.bands = np.array([d[i] for i in range(len(d))])
        # self.bands = self.lc_data.band_index

        # lambda_c needed for the color scatter estimation
        # lambda_c = self.training_dataset.lc_data['Wavelength']/(1+self.training_dataset.lc_data['ZHelio'])
        # idx_lam_c = np.array([lc.data['i'][0] for lc in self.training_dataset.lcs])
        # self.lambda_c = lambda_c[idx_lam_c]
        # self.lambda_c_red = self.color_law.reduce(self.lambda_c)
        #
        # TODO: where is this used ?
        # TODO: we may replace this by n_lc_meas > 0
        if self.training_dataset.lc_data is not None:
            tds = self.training_dataset
            ii = self.training_dataset.i_lc_first
            self.lambda_c = tds.lc_data.wavelength[ii] / (1+tds.lc_data.z[ii])
            self.lambda_c_red = self.color_law.reduce(self.lambda_c)

        # initialize degree of the spectral recalibration polynomial
        self._init_spectral_polynomial()

        # initialize model basis and model parameter vector
        # model component: bases, color law, and parameter vector
        self._init_bases(wl_grid, phase_grid,
                         wl_range, phase_range, basis_knots)

        # OK. For now, the color scatter and the calibration scatter
        # are explicitely instantiated in the model constructor
        self.error_snake_model = None
        self.calib_error_model = None
        self.color_scatter_model = None

        # finally, it is better to have it here.
        # NOTE: if there are spectra, the model must be evaluated for them.
        # even if they are marked as invalid
        self.recal_func = None
        if self.training_dataset.spec_data is not None:
            self.recal_func = \
                spectra.SpectrumRecalibrationPolynomials(self.training_dataset, self,
                                                 self.recalibration_degree)

        # initialize the global model parameters (M0, M1, CL)
        if init_from_salt2_file:
            self.init_from_salt2(init_from_salt2_file)
        else:
            self.pars = self.init_pars()
        # why ?
        # self.pars0 = self.pars.full.copy()

        # some of the classes gravitating around the model
        # the constraints for example, sometimes need to have
        # access to the index of some paramters in the full parameter
        # vector (all parameters released)
        self.all_pars_released = self.pars.copy()
        self.all_pars_released.release()

        # initialize the SN specific parameters (X0, X1, col, tmax)
        if init_from_training_dataset:
            self.init_from_training_dataset()

        # finalize the initialization of the error models
        # which need a fully operational parameter vector
        if self.calib_error_model is not None:
            self.calib_error_model.finalize()

        # we use a default model normalization
        # this is what guarantees that the SALT2.4 absolute mag
        # is -19.5 in the SWOPE::B band. Of course, this default
        # normalization may be changed explicitely, by calling
        # self.renorm(band_name=, Mb=, magsys='AB')
        self.norm = self.normalization(band_name=normalization_band_name,
                                       default_norm=1.01907246e-12)

        # grams
        self.init_grams_and_cc_grid()

        # and finally, prepare the computing units
        self.queue = self._init_computing_units()

        # clear the cache
        self.clear_cache()

        # if we need to run queue units separately,
        # could be useful to disable caching of the results
        self.disable_cache = False

    def register_variance_models(self, **kwargs):
        """register the (optional) error models

        Parameters
        ----------
        error_snake: salt2.variancemodels.ErrorSnake
          error snake
        calib: salt2.variancemodels.ExternalizedPrior
          calibration error model
        color_scatter: salt2.variancemodels.ExternalizedPrior
          color scatter model
        """
        self.error_snake_model = kwargs.get('error_snake', None)
        self.calib_error_model = kwargs.get('calib', None)
        self.color_scatter_model = kwargs.get('color_scatter', None)
        self.pars = self.init_pars()
        if self.color_scatter_model:
            self.color_scatter_model.update_pars(self.pars)
        if self.calib_error_model:
            self.calib_error_model.update_pars(self.pars)
        if self.color_scatter_model:
            self.color_scatter_model.update_pars(self.pars)

    def clone(self, training_dataset):
        """return a clone (same basis, same color law) for a different tds
        """
        ret = SALT2Like(training_dataset,
                        phase_range=self.phase_range,
                        wl_range=self.wl_range,
                        basis_knots=self.basis_knots,
                        basis_filter_knots=self.basis_filter_knots,
                        spectrum_recal_degree=self.spectrum_recal_degree,
                        normalization_band_name=self.normalization_band_name,
                        error_snake_model=self.error_snake_model,
                        calib_error_model=self.calib_error_model,
                        color_scatter_model=self.color_scatter_model)
                        #  calib_variance=self.calib_variance)
        for block_name in ['M0', 'M1', 'CL']:
            ret.pars[block_name].full[:] = self.pars[block_name].full[:]
        ret.norm = ret.normalization(band_name=ret.normalization_band_name)

        return ret

    def get_gram_dot_filter(self, band_name=None):
        """Used by the integral constraints
        """
        try:
            key = band_name if band_name is not None else self.normalization_band_name
            coeffs, _ = self.filter_db[key]
        except:
            key = 'standard::b'
            coeffs, _ = self.filter_db[key]
        return self.gram.dot(coeffs)

    def init_from_salt2_old(self, model_name, version=None):
        """Load & adapt the SALT2.4 global surfaces and color law.

        Load the coefficients of the SALT2.4 surfaces and color law. The
        definition of the model spline bases differs from the original
        definition of the SALT2.4 bases; so, we reproject the original SALT2
        surfaces on the model basis.

        Parameters
        ----------
        salt2_filename : str
        the classical salt2.npz filename containing the definition of the
        SALT2.4 bases, M0, M1 surfaces and color_law.
        """
        f = np.load(salt2_filename)
        phase_grid = f['phase_grid']
        wl_grid = f['wl_grid']
        basis = bspline.BSpline2D(wl_grid, phase_grid, x_order=4, y_order=4)

        if stick_to_original_model:
            # don't remember why we have this
            self.delta_phase = +0.7
            self.basis = basis
            self.pars = self.init_pars()
            self.pars['M0'].full[:] = f['M0'].T.ravel()
            self.pars['M1'].full[:] = f['M1'].T.ravel()
            self.pars['CL'].full[:] = f['CL_pars'][0:4]
            return

        # TODO: replace this with
        # xx = self.basis.bx.grid
        # yy = self.basis.by.grid
        xx = np.linspace(wl_grid[0], wl_grid[-1], basis.bx.nj)
        yy = np.linspace(phase_grid[0], phase_grid[-1], basis.by.nj)
        x,y = np.meshgrid(xx,yy)
        x,y = x.ravel(), y.ravel()
        jac = self.basis.eval(x,y).tocsr()
        factor = cholesky_AAt(jac.T, beta=1.E-6)

        # and initialize the parameters
        self.pars = self.init_pars()
        self.pars['M0'].full[:] = factor(jac.T * f['M0'])
        self.pars['M1'].full[:] = factor(jac.T * f['M1'])
        self.pars['CL'].full[:] = f['CL_pars'][0:4]
    
    def init_from_salt2(self, salt2_filename, stick_to_original_model=False):
        """Load & adapt the SALT2.4 global surfaces and color law from sncosmo.

        Load the coefficients of the SALT2.4 surfaces and color law. The
        definition of the model spline bases differs from the original
        definition of the SALT2.4 bases; so, we reproject the original SALT2
        surfaces on the model basis.

        Parameters
        ----------
        salt2_filename : str
        """
        salt2_source = sncosmo.get_source('salt2', version='2.4')
        
        #f = np.load(salt2_filename)
        #phase_grid = f['phase_grid']
        #wl_grid = f['wl_grid']
        
        #phase_grid = salt2_source._phase
        #wl_grid = salt2_source._wave
        
        
        phase_grid = np.linspace(self.phase_range[0], self.phase_range[1], self.n_ph)
        wl_grid = np.linspace(self.wl_range[0], self.wl_range[1], self.n_wl)
        basis = bspline.BSpline2D(wl_grid, phase_grid, x_order=4, y_order=4)
        
        phase_salt = salt2_source._phase
        wl_salt = salt2_source._wave
        
        w, p = np.meshgrid(wl_salt, phase_salt)
        
        sncosmo_scale = salt2_source._SCALE_FACTOR
        salt2_m0 = salt2_source._model['M0'](phase_salt, wl_salt).ravel() #/ sncosmo_scale
        salt2_m1 = salt2_source._model['M1'](phase_salt, wl_salt).ravel() #/ sncosmo_scale
        salt2_cl = np.array(salt2_source._colorlaw_coeffs)[::-1]
        
        jac = self.basis.eval(w.ravel(), p.ravel())
        factor = cholesky_AAt(jac.T, beta=1.E-1)

        # and initialize the parameters
        self.pars = self.init_pars()
        
        
        self.pars['M0'].full[:] = factor(jac.T * salt2_m0)
        self.pars['M1'].full[:] = factor(jac.T * salt2_m1)
        self.pars['CL'].full[:] = salt2_cl
    
    
    def reduce(self, wl):
        return self.color_law.reduce(wl)

    def _instantiate_error_model(self, error_model_type):
        if error_model_type is None:
            return None
        if type(error_model_type) == tuple:
            cls, args = error_model_type
            return cls(self, **args)
        if type(error_model_type) == type:
            return error_model_type(self)
        raise TypeError('arg should either be: None|(model_class, (args))|(model_class)')

    def init_grams_and_cc_grid(self):
        r"""Compute the :math:`\Lambda^{\mathrm{eff}}_{\ell q}` matrix (see definition above) on which the color
        law is evaluated.

        .. math::
             \bar{\lambda}_{\ell q} = \frac{\int \lambda^2 B_\ell(\lambda) B_q(\lambda)
             d\lambda}{\int \lambda B_\ell(\lambda) B_q(\lambda) d\lambda}

        Compute de grammian of order one, :math:`G` and two :math:`G2` of the model :

        .. math::
            G = \int \lambda B_\ell(\lambda) B_q(\lambda) d\lambda \\
            G2 = \int \lambda^2 B_\ell(\lambda) B_q(\lambda) d\lambda


        .. note::
             this method will be moved to the main model class (since we may work
             with one single :math:`\Lambda^{\mathrm{eff}}` matrix in a near future).
        """
        self.gram = get_gram(0., self.basis.bx, self.filter_basis, lambda_power=1)
        self.G2 = get_gram(0., self.basis.bx, self.filter_basis, lambda_power=2)

        gram = self.gram.tocoo()
        gram2 = self.G2.tocoo()
        assert(~np.any(gram.row-gram2.row) and ~np.any(gram.col-gram2.col))
        l_eff = gram2.data / gram.data
        self.L_eff = scipy.sparse.coo_matrix((l_eff, (gram.row, gram.col)), shape=gram.shape)

    def get_struct(self):
        """return the structure of the fit parameter vector

        In practice, the fit parameter vector is larger than just the model
        parameters: it also contains the parameters of the error models (error
        snake, color scatter, calibration, see e.g. variancemodels.py). The
        instantiation of the final fit parameter vector cannot therefore be
        performed by the model itself.  What the model can do, is to return the
        description of the fit parameter blocks it knows about.
        """
        nb_sne = self.training_dataset.nb_sne(valid_only=False)
        # nb_spectra = self.training_dataset.nb_spectra()
        # nb_passbands = len(self.bands)
        spec_recalibration_npars = self.recalibration_degree + 1
        nb_lightcurves = self.training_dataset.nb_lcs(valid_only=False)
        d = [('X0',   nb_sne),
             ('X1',   nb_sne),
             ('col',  nb_sne),
             ('tmax', nb_sne),
             ('M0',   self.basis.bx.nj * self.basis.by.nj),
             ('M1',   self.basis.bx.nj * self.basis.by.nj),
             ('CL',  4)]
        if self.training_dataset.spec_data is not None:
            d.append(('SpectrumRecalibration', spec_recalibration_npars.sum()))
        return d

    def init_pars(self):
        """instantiate a fit parameter vector

        Returns
        -------
        fp : nacl.lib.fitparameters.FitParameters
            Model parameters.
        """
        p_struct = self.get_struct()
        if self.calib_error_model is not None:
            p_struct.extend(self.calib_error_model.get_struct())
        if self.color_scatter_model is not None:
            p_struct.extend(self.color_scatter_model.get_struct())
        if self.error_snake_model is not None:
            p_struct.extend(self.error_snake_model.get_struct())
        fp = FitParameters(list(set(p_struct)))

        # minimal initialization
        fp['X0'].full[:] = 1.
        if 'SpectrumRecalibration' in fp._struct:
            assert self.recal_func is not None
            fp['SpectrumRecalibration'].full[:] = self.recal_func.init_pars()
        return fp

    def init_from_training_dataset(self):
        """load initial sn parameters from the training dataset
        """
        sn_data = self.training_dataset.sn_data
        self.pars['X0'].full[sn_data.sn_index] = sn_data.x0
        self.pars['X1'].full[sn_data.sn_index] = sn_data.x1
        self.pars['col'].full[sn_data.sn_index] = sn_data.col
        self.pars['tmax'].full[sn_data.sn_index] = sn_data.tmax

    def _init_spectral_polynomial(self):
        """instantiate the recalibration polynomials

        if a degree is specified, all spectra have the same degree. Since the
        degree of the polynomial must not exceed the number of Light curves :
        this degree is large odd number inferior to the numbers of light curves
        for this SN.
        """
        # note: the model is evaluated on the full dataset.
        # not just the valid data.
        nb_spectra = self.training_dataset.nb_spectra(valid_only=False)
        self.recalibration_degree = np.full(nb_spectra, self.spectrum_recal_degree)

    def _init_bases(self, wl_grid=None, phase_grid=None,
                    wl_range=(2000., 9000.),
                    phase_range=(-20., 50.), basis_knots=(127,20)):
        """
        Instantiate model bases
        """
        if wl_grid is None:
            assert wl_range is not None and basis_knots is not None
            logging.info('default regular grid in wavelength')
            wl_grid = np.linspace(self.wl_range[0], self.wl_range[1],
                                  self.n_wl)
        else:
            logging.info('user provided grid in wavelength')
        if phase_grid is None:
            assert phase_range is not None and basis_knots is not None
            logging.info('default regular grid in phase')
            phase_grid = np.hstack([np.linspace(self.phase_range[0],
                                                self.phase_range[1],
                                                self.n_ph)])
        else:
            logging.info('user provided grid in phase')
        self.basis = bspline.BSpline2D(wl_grid, phase_grid,
                                       x_order=4, y_order=4)

    @property
    def noise_models(self):
        """return a list of all the noise models contained by the model
        """
        ret = [self.calib_error_model,
               self.color_scatter_model,
               self.error_snake_model]
        return list(filter(lambda x: x is not None, ret))

    def normalization(self, band_name='SWOPE::B', Mb=-19.5, magsys='AB',
                      default_norm=None):
        """model normalization

        The SALT2Like normalization is set during training by the constraint
        on the integral of M0 at phase zero.

        Therefore, by default, we do not renormalize the model during
        evaluation.
        """
        try:
            tq, filter_basis = self.filter_db[band_name]
        except:
            tq, filter_basis = self.filter_db['standard::b']
        phase_eval = self.basis.by.eval(np.array([0. + self.delta_phase])).tocsr()
        gram = get_gram(z=0., model_basis=self.basis.bx,
                        filter_basis=filter_basis, lambda_power=1)
        surface_0 = self.pars['M0'].full.reshape(len(self.basis.by), -1)
        # evaluate the integral of the model in the specified band
        self.int_M0_phase_0 = phase_eval.dot(surface_0.dot(gram.dot(tq)))

        # AB flux in the specified band
        try:
            ms = MagSys(magsys)
            zp = ms.ZeroPoint(self.filter_db.transmission_db[band_name])
        except:
            m_AB = sncosmo.ABMagSystem()
            band = sncosmo.get_bandpass('standard::b')
            zp = -2.5*np.log10(m_AB.zpbandflux(band))
            
        self.int_ab_spec = 10**(0.4 * zp)

        # normalization quantities
        self.flux_at_10pc = np.power(10., -0.4 * (Mb-zp))
        self.flux_at_10Mpc = np.power(10., -0.4 * (Mb+30.-zp))

        if default_norm is not None:
            return default_norm

        return self.flux_at_10Mpc / self.int_M0_phase_0
        #return 1.

    def renorm(self, band_name='SWOPE::B', Mb=-19.5, magsys='AB'):
        """Adjust the model normalization
        """
        # explicitely recompute a normalization
        self.norm = self.normalization(band_name, Mb, magsys,
                                       default_norm=None)

    def _init_computing_units(self):
        """Prepare and queue all the computing units

        The model evaluation is deferred to two kinds of so-called
        'eval units'.

            - Photometric eval units, which compute the light curves
              (i.e. the integral of the SALT2 surfaces). In our current
              design, a Photometric eval unit predicts the lightcurve
              for a given SN in a given band.
            - Spectroscopic eval units, which predict SN spectra.
              In our current design, a spectroscopic eval unit
              predicts one spectrum for a given SN.

        All units are grouped in a single execution queue.

        .. note:: we designed things like this because we had some
             hope that we could gain in speed from using explicit
             parallelism (using joblib or some similar framework).  In
             practice, it seems that the overheads from using python
             parallelism at this level are high (because of the GIL)
             and that we gain much more by using numpy/scipy implicit
             multithreading. We stick to this design because it makes
             things clearer in fact.

        """
        # n_lc_meas, n_spec_meas, n_spectrophot_meas = \
        #     self.training_dataset.nb_meas(valid=True, split_by_type=True)
        queue = []
        if self.training_dataset.lc_data is not None:
            for lc in self.training_dataset.lcs:
                queue.append(lightcurves.LightcurveEvalUnitTz(lc, self))
        if self.training_dataset.spec_data is not None:
            queue.append(spectra.SpectrumEvalUnitFast(self.training_dataset, self))
        if self.training_dataset.spectrophotometric_data is not None:
            queue.append(spectra.SpectroPhotoEvalUnit(self.training_dataset, self))
        return queue

    def update_computing_units(self, ilc=None, spec=False):
        """
        Recreate the model queue as a function of the wanted LCs and spectra.

        Parameters
        -------
        ilc : None or int
            If None all light curve are evaluated, else only teh desired one.
        spec : bool
            Whether Spectra need to be evaluated.

        Returns
        -------
        queue : list
            The new queue
        """
        queue = []
        if ilc is not None:
            for lc in [self.training_dataset.lcs[ilc]]:
                queue.append(lightcurves.LightcurveEvalUnitTz(lc, self))
        if spec:
            queue.append(spectra.SpectrumEvalUnitFast(self.training_dataset, self))
        return queue

    def clear_cache(self):
        """
        Clear model evaluation, derivatives and timing.
        """
        self.val = []
        self.jacobian_i = []
        self.jacobian_j = []
        self.jacobian_val = []
        self.timing = []

    def precompute_color_law(self, cl_pars=None, jac=False):
        """
        Precompute the color law for the photometric data.

        Parameters
        -------
        cl_pars : None or numpy.array
            Color law parameters
        jac : bool
            If derivatives are needed.
        """
        if cl_pars is None:
            cl_pars = self.pars['CL'].full

        polynome_color_law, jacobian_color_law = \
            self.color_law(self.L_eff.data, cl_pars,
                           jac=jac)  # return_jacobian_as_coo_matrix=False)
        self.polynome_color_law = \
            scipy.sparse.csr_matrix((polynome_color_law,
                                     (self.L_eff.row, self.L_eff.col)),
                                    shape=self.L_eff.shape)
        if jac:
            self.jacobian_color_law = []
            _, n = jacobian_color_law.shape
            for i in range(n):
                jacobian_cl = \
                    scipy.sparse.csr_matrix((jacobian_color_law[:, i],
                                             (self.L_eff.row, self.L_eff.col)), shape=self.L_eff.shape)
                self.jacobian_color_law.append(jacobian_cl)

    def get_restframe_phases(self, data):
        """
        Given a chunk of data, return the restframe phases

        Returns
        -------
        numpy.array
            data SN-restframe phase
        """
        sn_index = data.sn_index
        tmax = self.pars['tmax'].full[sn_index]
        return (data.mjd - tmax) / (1.+data.z)

    def __call__(self, p, jac=False, plotting=False, ilc_plot=None, spec_plot=False):
        """Evaluate the model for the parameter set p

        In practice, loop over the queue and run each eval unit.
        Assemble the results into one single vector and (optionally)
        one single jacobian matrix.

        Parameters
        -------
        p : numpy.array
            Vector containing the free parameters only.
        jac : bool
            Whether it return the jacobian matrix.

        plotting : bool
            Plotting need to evaluated only some data [plotting needs]
        ilc_plot : list
            Light curves to be evaluated [plotting needs]
        spec_plot : bool
            Spectra to be evaluated [plotting needs]

        Returns
        -------
        val : numpy.array
            model results
        jacobian : scipy.sparse.csr_matrix
            jacobian matrix (if jac is true)
        """
        # update pars
        #self.pars.free = p.free

        # pre-evaluate the color law (shared between all light curves)
        self.precompute_color_law(jac=jac)

        # precompute the values of the phase basis on the full dataset
        # significantly faster that way
        if self.training_dataset.lc_data is not None:
            restframe_phases = self.get_restframe_phases(self.lc_data)
            self.phase_eval = self.basis.by.eval(restframe_phases + self.delta_phase).tocsr()
            if jac:
                self.dphase_dtmax = self.basis.by.deriv(restframe_phases + self.delta_phase).tocsr()
            else:
                self.dphase_dtmax = None

        self.clear_cache()

        # queue all Eval Unit for minimization or for plotting
        if plotting:
            queue = self.update_computing_units(ilc=ilc_plot, spec=spec_plot)
        else:
            queue = self.queue

        # now, we are ready to loop over the eval units
        for q in queue:
            q(jac)
        val = np.hstack(self.val)

        if not jac:
            return val

        # collect and assemble the results
        # n_data = len(self.training_dataset.lc_data)  + len(self.training_dataset.spec_data)
        n_data = self.training_dataset.nb_meas(valid_only=False)

        # nrl: I don't know what this is for. Commented out.
        # fit_spec = False
        # for ique in self.queue:
        #     if ique.data['spec_id'].mean() != -1:
        #         fit_spec = True

        # if fit_spec:
        #     n_data += len(self.training_dataset.spec_data)

        logging.debug('computing derivatives: hstack...')
        n = len(p.free)
        i = np.hstack(self.jacobian_i)
        j = np.hstack(self.jacobian_j)
        v = np.hstack(self.jacobian_val)
        logging.debug('building coo_matrix...')
        idx = j >= 0  # self.pars.indexof(j) >= 0
        jacobian = scipy.sparse.coo_matrix((v[idx], (i[idx], j[idx])), shape=(n_data, n))
        logging.debug('ok, done.')
        return val, jacobian


_G, _G2 = {}, {}

def get_gram(z, model_basis, filter_basis, lambda_power=1):
    """
    Calculate the grammian of to spline basis.

    The grammian :math:`G` of order :math:`N`, to basis :math:`B_0` (define on wavelength,
    in our case surfaces wavelength basis) and
    :math:`B_1` (define on SN-restframe wavelength, filter basis) is defined as :

    .. math::
        G = \\int \\lambda^N B_0(\\lambda) B_q(\\lambda (1+z)) d\\lambda

    Parameters
    -------
    z : numpy.array
        Vector containing the data redshift.
    model_basis : nacl.lib.bspline.BSpline
        Wavelength basis.
    filter_basis : nacl.lib.bspline.BSpline
        Filter basis defined common to all the bands.
    lambda_power : int
        Grammian order.

    Returns
    -------
    gram : scipy.sparse.csc_matrix
        Grammian
    """
    global _G
    global _G2
    key = (z, model_basis, filter_basis, lambda_power)

    gram = _G.get(key, None)
    if gram is None:
        gram = bspline.lgram(model_basis, filter_basis, z=z, lambda_power=lambda_power).tocsc()
        _G[key] = gram

    return gram





"""

"""

import numpy as np
import scipy.sparse as sparse
from ....lib.fitparameters import FitParameters
from ....lib import bspline


class Model:
    """A more realistic 2D lightcurve model
    """

    def __init__(self, data, grid_x1=None, grid_x2=None):
        self.data = data
        self.nsn = len(data.sn_map)
        if grid_x1 is not None:
            self.grid_x1 = grid_x1
            self.grid_x2 = grid_x2
        else:
            self.grid_x1 = np.linspace(self.data.x1.min() - 5., self.data.x1.max() + 5., 25)
            self.grid_x2 = np.linspace(self.data.x2.min() - 5., self.data.x2.max() + 5., 25)
        self.basis = bspline.BSpline2D(self.grid_x1, self.grid_x2, x_order=4, y_order=4)


    def get_struct(self):
        return [('x0', self.nsn), ('stretch', self.nsn), ('tmax', self.nsn),
                ('theta', len(self.basis))]

    def init_pars(self, pars=None):
        if pars is None:
            pars = FitParameters(self.get_struct())
        n = len(pars['theta'].full)
        pars['theta'].full[:] = np.random.uniform(-1., 1., len(self.basis))
        pars['x0'].full[:] = np.ones(self.nsn)
        pars['stretch'].full[:] = np.zeros(self.nsn)
        pars['tmax'].full[:] = 0.
        return pars

    def __call__(self, pars, jac=False):
        """evaluate the model and (optionally) its derivatives

        Parameters
        ----------
        pars: (FitParameters)
          model parameter vector
        jac: (bool)
          whether to evaluate or not the jacobian

        Returns
        -------
        ndarray of floats if jac is False
        ndarray, sparse.csr_matrix: if jac is True
        """
        x0   = pars['x0'].full[self.data.sn_index]
        stretch = pars['stretch'].full[self.data.sn_index]
        tmax = pars['tmax'].full[self.data.sn_index]

        dx1 = self.data.x1 - tmax
        dx2 = self.data.x2 - tmax
        ph1 = (1. + stretch) * dx1
        ph2 = (1. + stretch) * dx2
        #P1, P2 = np.meshgrid(ph1.to_numpy(), ph2.to_numpy())
        #J = self.basis.eval(P1.ravel(), P2.ravel())
        J = self.basis.eval(ph1.to_numpy(), ph2.to_numpy())
        mod = J @ pars['theta'].full
        vals = x0 * mod

        if not jac:
            return vals

        # model derivatives
        N = len(vals)
        i = np.arange(N)
        ii, jj, data = [], [], []
        # dmodel_dx0
        ii.append(i)
        jj.append(np.full(N, pars['x0'].indexof(self.data.sn_index)))
        data.append(mod)

        # dmodel_dstretch
        J_dmodel1, J_dmodel2 = self.basis.gradient(ph1.to_numpy(), ph2.to_numpy())
        dmod1 = J_dmodel1 @ pars['theta'].full
        dmod2 = J_dmodel2 @ pars['theta'].full
        ii.append(i)
        jj.append(pars['stretch'].indexof(self.data.sn_index))
        data.append(x0 * (dmod1 * dx1 + dmod2 * dx2))

        # dmodel_dtmax
        ii.append(i)
        jj.append(pars['tmax'].indexof(self.data.sn_index))
        data.append(-x0 * (dmod1 + dmod2) * (1. + stretch))

        # dmodel_dtheta
        ii.append(J.row)
        jj.append(pars['theta'].indexof(J.col))
        data.append(x0[J.row] * J.data)

        # build the matrix
        i = np.hstack(ii)
        j = np.hstack(jj)
        data = np.hstack(data)
        idx = j>=0

        n_free_pars = len(pars.free)
        J = sparse.coo_matrix((data[idx], (i[idx], j[idx])), shape=(N,n_free_pars))

        return vals, J
        
        
class ModelwithCalibScatter:
    """A more realistic 2D lightcurve model
    """

    def __init__(self, data, grid_x1=None, grid_x2=None):
        self.data = data
        self.nsn = len(data.sn_map)
        self.nbands = len(self.data.band_set)
        if grid_x1 is not None:
            self.grid_x1 = grid_x1
            self.grid_x2 = grid_x2
        else:
            self.grid_x1 = np.linspace(self.data.x1.min() - 5., self.data.x1.max() + 5., 25)
            self.grid_x2 = np.linspace(self.data.x2.min() - 5., self.data.x2.max() + 5., 25)
        self.basis = bspline.BSpline2D(self.grid_x1, self.grid_x2, x_order=4, y_order=4)


    def get_struct(self):
        return [('x0', self.nsn), ('stretch', self.nsn), ('tmax', self.nsn),
                ('theta', len(self.basis)),
                ('eta', self.nbands)]

    def init_pars(self, pars=None):
        if pars is None:
            pars = FitParameters(self.get_struct())
        n = len(pars['theta'].full)
        pars['theta'].full[:] = np.random.uniform(-1., 1., len(self.basis))
        pars['x0'].full[:] = np.ones(self.nsn)
        pars['stretch'].full[:] = np.zeros(self.nsn)
        pars['tmax'].full[:] = 0.
        pars['eta'].full[:] = 0.
        return pars

    def __call__(self, pars, jac=False):
        """evaluate the model and (optionally) its derivatives

        Parameters
        ----------
        pars: (FitParameters)
          model parameter vector
        jac: (bool)
          whether to evaluate or not the jacobian

        Returns
        -------
        ndarray of floats if jac is False
        ndarray, sparse.csr_matrix: if jac is True
        """
        x0   = pars['x0'].full[self.data.sn_index]
        stretch = pars['stretch'].full[self.data.sn_index]
        tmax = pars['tmax'].full[self.data.sn_index]
        
        #colorscatter
        cal = 1. + pars['eta'].full[self.data.band_index]

        dx1 = self.data.x1 - tmax
        dx2 = self.data.x2 - tmax
        ph1 = (1. + stretch) * dx1
        ph2 = (1. + stretch) * dx2
        #P1, P2 = np.meshgrid(ph1.to_numpy(), ph2.to_numpy())
        #J = self.basis.eval(P1.ravel(), P2.ravel())
        J = self.basis.eval(ph1.to_numpy(), ph2.to_numpy())
        mod = J @ pars['theta'].full
        vals = x0 * mod * cal

        if not jac:
            return vals

        # model derivatives
        N = len(vals)
        i = np.arange(N)
        ii, jj, data = [], [], []
        # dmodel_dx0
        ii.append(i)
        jj.append(np.full(N, pars['x0'].indexof(self.data.sn_index)))
        data.append(mod * cal)

        # dmodel_dstretch
        J_dmodel1, J_dmodel2 = self.basis.gradient(ph1.to_numpy(), ph2.to_numpy())
        dmod1 = J_dmodel1 @ pars['theta'].full
        dmod2 = J_dmodel2 @ pars['theta'].full
        ii.append(i)
        jj.append(pars['stretch'].indexof(self.data.sn_index))
        data.append(x0 * (dmod1 * dx1 + dmod2 * dx2) * cal)

        # dmodel_dtmax
        ii.append(i)
        jj.append(pars['tmax'].indexof(self.data.sn_index))
        data.append(-x0 * (dmod1 + dmod2) * (1. + stretch) * cal)

        # dmodel_dtheta
        ii.append(J.row)
        jj.append(pars['theta'].indexof(J.col))
        data.append(x0[J.row] * J.data * cal[J.row])
        
        # dmodel_deta
        ii.append(i)
        jj.append(pars['eta'].indexof(self.data.band_index))
        data.append(x0 * mod)

        # build the matrix
        i = np.hstack(ii)
        j = np.hstack(jj)
        data = np.hstack(data)
        idx = j>=0

        n_free_pars = len(pars.free)
        J = sparse.coo_matrix((data[idx], (i[idx], j[idx])), shape=(N,n_free_pars))

        return vals, J
        
class LightCurve2DCalibScatter:
    """A more realistic 2D lightcurve model with calibration errors and a color parameter
    """

    def __init__(self, data, grid_x1=None, grid_x2=None, pivot_wl=5000.):
        self.data = data
        self.nsn = len(data.sn_map)
        self.nbands = len(self.data.band_set)
        self.pivot_wl = pivot_wl
        if grid_x1 is not None:
            self.grid_x1 = grid_x1
            self.grid_x2 = grid_x2
        else:
            self.grid_x1 = np.linspace(self.data.x1.min() - 5., self.data.x1.max() + 5., 25)
            self.grid_x2 = np.linspace(self.data.x2.min() - 5., self.data.x2.max() + 5., 25)
        self.basis = bspline.BSpline2D(self.grid_x1, self.grid_x2, x_order=4, y_order=4)


    def get_struct(self):
        return [('x0', self.nsn), ('stretch', self.nsn), 
                ('tmax', self.nsn), ('color', self.nsn),
                ('theta', len(self.basis)),
                ('eta', self.nbands)]

    def init_pars(self, pars=None):
        if pars is None:
            pars = FitParameters(self.get_struct())
        n = len(pars['theta'].full)
        pars['theta'].full[:] = np.random.uniform(-1., 1., len(self.basis))
        pars['x0'].full[:] = np.ones(self.nsn)
        pars['stretch'].full[:] = np.zeros(self.nsn)
        pars['tmax'].full[:] = 0.
        pars['color'].full[:] = 0.
        pars['eta'].full[:] = 0.
        return pars

    def __call__(self, pars, jac=False):
        """evaluate the model and (optionally) its derivatives

        Parameters
        ----------
        pars: (FitParameters)
          model parameter vector
        jac: (bool)
          whether to evaluate or not the jacobian

        Returns
        -------
        ndarray of floats if jac is False
        ndarray, sparse.csr_matrix: if jac is True
        """
        x0   = pars['x0'].full[self.data.sn_index]
        stretch = pars['stretch'].full[self.data.sn_index]
        tmax = pars['tmax'].full[self.data.sn_index]
        zz = self.data.zz
        # rather simple color law - just for tests
        dwl = ((self.data.wl - self.pivot_wl) / self.pivot_wl).to_numpy()
        cl = 1. + pars['color'].full[self.data.sn_index] * dwl
        
        #colorscatter
        cal = 1. + pars['eta'].full[self.data.band_index]

        dx1 = self.data.x1 - tmax
        dx2 = self.data.x2 - tmax
        ph1 = (1. + stretch) * dx1 / zz
        ph2 = (1. + stretch) * dx2 / zz
        #P1, P2 = np.meshgrid(ph1.to_numpy(), ph2.to_numpy())
        #J = self.basis.eval(P1.ravel(), P2.ravel())
        J = self.basis.eval(ph1.to_numpy(), ph2.to_numpy())
        mod = J @ pars['theta'].full
        vals = x0 * mod * cal * cl

        if not jac:
            return vals

        # model derivatives
        N = len(vals)
        i = np.arange(N)
        ii, jj, data = [], [], []
        # dmodel_dx0
        ii.append(i)
        jj.append(np.full(N, pars['x0'].indexof(self.data.sn_index)))
        data.append(mod * cl * cal)

        # dmodel_dstretch
        J_dmodel1, J_dmodel2 = self.basis.gradient(ph1.to_numpy(), ph2.to_numpy())
        dmod1 = J_dmodel1 @ pars['theta'].full
        dmod2 = J_dmodel2 @ pars['theta'].full
        ii.append(i)
        jj.append(pars['stretch'].indexof(self.data.sn_index))
        data.append(x0 * (dmod1 * dx1 + dmod2 * dx2) * cl * cal / zz)
        
        # dmodel_dcolor
        ii.append(i)
        jj.append(pars['color'].indexof(self.data.sn_index))
        data.append(x0 * mod * dwl * cal)

        # dmodel_dtmax
        ii.append(i)
        jj.append(pars['tmax'].indexof(self.data.sn_index))
        data.append(-x0 * (dmod1 + dmod2) * (1. + stretch) * cl * cal / zz)

        # dmodel_dtheta
        ii.append(J.row)
        jj.append(pars['theta'].indexof(J.col))
        data.append(x0[J.row] * J.data * cl[J.row] * cal[J.row])
        
        # dmodel_deta
        ii.append(i)
        jj.append(pars['eta'].indexof(self.data.band_index))
        data.append(x0 * mod * cl)

        # build the matrix
        i = np.hstack(ii)
        j = np.hstack(jj)
        data = np.hstack(data)
        idx = j>=0

        n_free_pars = len(pars.free)
        J = sparse.coo_matrix((data[idx], (i[idx], j[idx])), shape=(N,n_free_pars))

        return vals, J



#!/usr/bin/env python3

import logging

import numpy as np
import matplotlib.pyplot as plt

from lemaitre import bandpasses
from nacl.models import helpers
from nacl.models import salt2
from nacl.models.salt2 import lightcurves
from nacl.models.salt2 import spectra
from nacl import TrainingDataset


def load(filename='tds_test_deriv', keep=None):
    """
    """
    fl = bandpasses.get_filterlib()

    if 'hd5' in filename:
        tds = TrainingDataset.read_hdf(filename)
    else:
        tds = TrainingDataset.read_parquet(filename, filterlib=fl)

    if keep is not None:
        sne = tds.sn_data.sn_set.tolist()
        for k in keep:
            sne.remove(k)
        tds.kill_sne(sne)
        tds.compress()

    return tds


def test_kron(filename='compressed_lm2'):

    tds = load(filename)
    model = salt2.SALT2Like(tds)
    pars = model.init_pars()
    ll = lightcurves.LightcurveEvalUnit(model)
    F = ll.meas_filter_projections
    restframe_phase = (tds.lc_data.mjd - pars['tmax'].full[tds.lc_data.sn_index]) / (1. + tds.lc_data.z)
    ph_basis = ll.model.basis.by
    J = ph_basis.eval(restframe_phase)

    return J, F


def test_deriv(model, pars, block_name='M0'):
    """
    """
    #       pars = model.init_pars()
    pars.fix()
    pars.release(block_name)

    J, Jn = helpers.check_grad(model, pars, dx=1.E-7)
    print(J.shape, Jn.shape)
    J = J.ravel()
    Jn = Jn.ravel()
    idx = (np.abs(J) > 0) | (np.abs(Jn) > 0)

    fig, axes = plt.subplots(nrows=2, ncols=2, figsize=(10,10))
    fig.suptitle(f'derivatives: {block_name}')
    axes[0,0].plot(J[idx], Jn[idx], 'k.')
    axes[0,0].set_xlabel('$\partial_{true}$')
    axes[0,0].set_ylabel('$\partial_{num}$')

    axes[0,1].plot((J[idx] - Jn[idx])/Jn[idx], 'k.')
    axes[0,1].set_xlabel('i')
    axes[0,1].set_ylabel('$(\partial_{true} - \partial_{num}) / \partial_{num}$')

    axes[1,0].plot(J[idx], (J[idx] - Jn[idx])/Jn[idx], 'k.')
    axes[1,0].set_xlabel('$\partial_{true}$')
    axes[1,0].set_ylabel('$(\partial_{num} - \partial_{true})/\partial_{num}$')

    axes[1,1].plot(J[idx] - Jn[idx], 'k.')
    axes[1,1].set_xlabel('$i$')
    axes[1,1].set_ylabel('$\partial_{true} - \partial_{num}$')

    return J, Jn


if __name__ == '__main__':
    tds = load()
    model = salt2.SALT2Like(tds)
    pars = model.init_pars()
    # wl = np.linspace(2200., 9200., 250)

    ll = lightcurves.LightcurveEvalUnit(model)

    for block_name in ['X0', 'X1', 'col', 'tmax', 'M0', 'M1', 'CL']:
        logging.info(f'testing: {block_name}')
        test_deriv(ll, pars, block_name)

    pars.fix()
    pars.release('M0')
    J, Jn = helpers.check_grad(ll, pars, dx=1.E-7)
    fig, axes = plt.subplots(nrows=1, ncols=3, figsize=(16,5),
                             sharex=True, sharey=True)
    fig.suptitle('M0')
    axes[0].imshow(J, aspect='auto', vmin=-0.1, vmax=0.1)
    axes[0].set_title('analytical')
    axes[1].imshow(Jn, aspect='auto', vmin=-0.1, vmax=0.1)
    axes[1].set_title('numerical')
    c = axes[2].imshow((J-Jn)/Jn, aspect='auto', vmin=-1., vmax=1.)
    plt.colorbar(c)
    axes[2].set_title('(J-Jn)/Jn')

    pars.fix()
    pars.release('M1')
    J, Jn = helpers.check_grad(ll, pars)
    fig.suptitle('M1')
    fig, axes = plt.subplots(nrows=1, ncols=3, figsize=(16,5),
                             sharex=True, sharey=True)
    axes[0].imshow(J, aspect='auto', vmin=-0.1, vmax=0.1)
    axes[0].set_title('analytical')
    axes[1].imshow(Jn, aspect='auto', vmin=-0.1, vmax=0.1)
    axes[1].set_title('numerical')
    c = axes[2].imshow((J-Jn)/Jn, aspect='auto', vmin=-1., vmax=1.)
    plt.colorbar(c)
    axes[2].set_title('(J-Jn)/Jn')




import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
import pandas
from ....lib import bspline


def gaussian(x, a, sigma):
    """
    Gaussian function *2
    """
    norm = a / (np.sqrt(2. * np.pi) * sigma)
    xx = x / sigma
    return norm * np.exp(-0.5 * xx**2)


class ToyDataGenerator:

    def __init__(self, xmin=-10., xmax=10., yerr=0.01, a=1., sigma=3.,
                 n=100, delta=2., bsize=10, order=4, error_pedestal=0.025):
        self.xmin, self.xmax = xmin, xmax
        self.yerr = yerr
        self.error_pedestal = error_pedestal
        if self.error_pedestal is not None:
            self.sig_noise = np.sqrt(self.yerr**2 + self.error_pedestal**2)
        else:
            self.sig_noise = yerr
        self.a, self.sigma = a, sigma
        self.n = n
        self.delta = delta
        self.basis = bspline.BSpline(np.linspace(xmin-delta, xmax+delta, bsize), order=order)
        self.error_pedestal=error_pedestal

    def generate(self):
        """
        generate a realization of a dataset
        """
        x = np.random.uniform(self.xmin, self.xmax, size=self.n)
        y = gaussian(x, self.a, self.sigma)
        yerr = np.full(self.n, self.yerr)
        sig_noise = np.full(self.n, self.sig_noise)
        noise = np.random.normal(loc=0., scale=sig_noise)
        self.x, self.y_true, self.yerr, self.noise = x, y, yerr, noise
        self.y = self.y_true + self.noise
        return self.x, self.y + self.noise, self.yerr

    def get_data(self):
        if not hasattr(self, 'x'):
            self.generate()
        dd = pandas.DataFrame({'xx': self.x,
                               'y': self.y,
                               'yerr': self.yerr,
                               'bads': np.zeros(len(self.yerr)).astype(bool),
                               })
        return dd

    def fit(self, beta=1.E-6):
        """a fit of the data -- the simplest possible
        """
        J = self.basis.eval(self.x)
        w = 1. / self.yerr**2
        W = dia_matrix((w, [0]), shape=(self.n, self.n))
        H = J.T @ W @ J
        B = J.T @ W @ self.y
        fact = cholmod.cholesky(H, beta=beta)
        self.theta = fact(B)

    def plot(self, color='b', marker='.'):
        """plot the data and the model
        """
        fig = plt.figure()
        gs = GridSpec(2, 1, height_ratios=[4,2])

        ax1 = fig.add_subplot(gs[0])
        ax1.errorbar(self.x, self.y, yerr=self.yerr, ls='', color=color, marker=marker, label='data', zorder=0.5)
        ax1.set_ylabel('model \& data')
        ax2 = fig.add_subplot(gs[1], sharex=ax1)
        ax2.set_xlabel('x')
        ax2.set_ylabel('residuals')

        # pl.grid(ls=':')
        # pl.plot(self.basis.grid, np.zeros(len(self.basis.grid)), 'k|')
        if hasattr(self, 'theta'):
            xx = self.basis._grid.copy()
            JJ = self.basis.eval(xx)
            ax1.plot(xx, JJ@self.theta, ls='', color='red', marker='.', alpha=1., label='nodes', zorder=10., linewidth=2)

            xx = np.linspace(self.xmin-self.delta, self.xmax+self.delta, 200)
            JJ = self.basis.eval(xx)
            ax1.plot(xx, JJ@self.theta, 'r-', label='model', zorder=2.)

            JJ = self.basis.eval(self.x)
            ax2.errorbar(self.x, self.y-JJ @ self.theta, yerr=self.yerr, ls='', marker='.', color='k')

        plt.legend(loc='best')



def plot_model_and_data(model, pars, data, title='', reg_lambda=None):
    """
    """
    fig = plt.figure()
    gs = GridSpec(3, 1, height_ratios=[4,2,2])

    ax1 = fig.add_subplot(gs[0])
    ax1.errorbar(data.xx, data.y, yerr=data.yerr, ls='', marker='.')
    xx = np.linspace(model.basis.grid.min(), model.basis.grid.max(), 200)
    J = model.basis.eval(xx)
    ax1.plot(xx, J @ pars['theta'].full, 'r-', zorder=10)
    plt.setp(ax1.get_xticklabels(), visible=False)

    ax1.set_ylabel('model \& data')
    ax2 = fig.add_subplot(gs[1], sharex=ax1)
    res = data.y-model(pars)
    ax2.errorbar(data.xx, res, yerr=data.yerr, ls='', marker='.')
    ax2.set_xlabel('x')
    ax2.set_ylabel('residuals')
    plt.setp(ax2.get_xticklabels(), visible=False)

    ax3 = fig.add_subplot(gs[2], sharex=ax1)
    pars = pars.copy()
    xx = np.linspace(model.basis.grid.min(), model.basis.grid.max(), 500)
    J = model.basis.eval(xx)
    for i in range(len(pars['theta'].full)):
        pars['theta'].full[:] = 0.
        pars['theta'].full[i] = 3 * res.std()
        if reg_lambda is not None:
            color = plt.cm.jet(int(reg_lambda[i]/reg_lambda.max() * 256))
            # color = 'r' if reg_lambda[i]>0. else 'b'
        else:
            color = plt.cm.jet(int(i * 256/len(pars['theta'].full)))
        ax3.plot(xx, J@pars['theta'].full, ls='-', color=color)
    ax3.set_xlabel('x')

    plt.subplots_adjust(hspace=0.05)
    fig.suptitle(title)



import numpy as np
import pylab as pl
from nacl.models.salt import SALT2Like
from nacl import dataset
from nacl.simulations.fullsim import sngen

import helpers

tds, model = helpers.generate_dataset(2, seed=42)


def test_X0_definition():
    """
    Here, we check the normalization of the model.
    Our goal is to clarify the normalization of the model, while
    keeping the overall values of the X0's and the overall values
    of the M0,M1 surfaces not too far away from the original model.
    """
    gen = sngen.DVDzSNGenerator(z_range=(0.001, 0.6))
    sample = gen.generate_sample(nsn=10000)
    jla = dataset.read_hdf('jla.hd5')
    zz = np.linspace(0.001, 0.6, 500)

    # JLA versus d_L(z) values
    pl.figure()
    pl.semilogy(jla.sn_data.z, jla.sn_data.x0, 'k.',
                label="original JLA X0's")
    pl.semilogy(zz, 100. / gen.cosmo.luminosity_distance(zz)**2, 'r--',
                label='$100/d_L^2(z)$')
    pl.semilogy(zz,  80. / gen.cosmo.luminosity_distance(zz)**2, 'c--',
                label='$80/d_L^2(z)$')
    pl.legend(loc='best')
    pl.xlabel('redshift')
    pl.ylabel('X0')
    pl.title('Comparison between the JLA $X_0$ values and $d_L(z)$' )

    # generator versus d_L(z) values
    pl.figure()
    pl.semilogy(sample.z, sample.x0, 'k.',
                label='simulated sample')
    pl.semilogy(zz, 100. / gen.cosmo.luminosity_distance(zz)**2, 'r--',
                label='$100/d_L(z)^2$')
    pl.legend(loc='best')
    pl.xlabel('redshift')
    pl.ylabel('X0')
    pl.title('Comparison between the generator $X_0$ values and $d_L(z)$' )

    return gen, jla





import logging
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                    level=logging.INFO)

import time 
import os
import os.path as op
import numpy as np
import scipy.sparse
import sksparse.cholmod

import croaks
import saunerie
from saunerie.fitparameters import FitParameters

from salt3.minimizers import minimize_nr_var_quad#, minimize_nr_qr
import salt3.simulations.generator
import salt3.models.models

import salt3.models.variancemodels
import salt3.models.regularizations
import salt3.models.constraints

import salt3.plotting.plots

from helpers import locate_test_datafile, check_grad, check_grad_var
##############################################
import pylab as pl
pl.ion()

bands = ['B','V', 'g', 'r']#, 'u'] # 
def fit_1dsplinemodelmultiband_same_lcs(N=100, bands  = bands,
                                        minimize=minimize_nr_var_quad, mu_regularization=10.,
                                        plot=False, seed=None):
    """
    """
    lc_data = salt3.simulations.generator.generate_lc_multiband(N, bands,
                                                    tmax_range=(0., 0.),
                                                    sig_range=(0., 0.),
                                                    norm_range=(1000., 1000.),
                                                    seed=seed) 
    
    model = salt3.models.models.MultiBandsModel(lc_data, rng_spline = [-100, 100], grid_size=250)
    f = salt3.models.models.VarModelResiduals(lc_data, model)
    cons = salt3.models.constraints.NonIntegralConstraints_4MultibandModel(f.pars, model, band_tmax='B')    
    reg = salt3.models.regularizations.SplineParamsRegularizationMultiband(f.pars, bands = model.bands_unique,
                                                           mu=mu_regularization)

    # initialization fit (spline parameters only)
    for bd in model.bands_unique:
        f.pars[f"flux_{bd.decode('UTF-8')}"].full[:] = 1000
        f.pars[f"theta_{bd.decode('UTF-8')}"].full[:] = 1.E-6
        f.pars[f"flux_{bd.decode('UTF-8')}"].fix()
    f.pars['tmax'].fix()
    f.pars['s'].fix()
    cons = salt3.models.constraints.NonIntegralConstraints_4MultibandModel(f.pars, model,
                                                               band_tmax='B')
    reg = salt3.models.regularizations.SplineParamsRegularizationMultiband(f.pars,
                                                           bands = model.bands_unique,
                                                           mu=mu_regularization)
    x = minimize(f, f.pars.free, reg=reg, cons=cons)
    f.pars.free = x[0]
    
    # real fit, with all parameters free 
    f.pars.release()
    for bd in model.bands_unique:
        f.pars[f"flux_{bd.decode('UTF-8')}"].full[:] = 1000

    #f.pars['flux'].full[:] = 10000
    x = minimize(f, f.pars.free, cons=cons, reg=reg)
    f.pars.free = x[0]

    if plot:
        import pylab as pl
        import matplotlib as mpl
        pl.figure()
        gs = mpl.gridspec.GridSpec(4,1)
        ax = pl.subplot(gs[0:3,0])
        dif = pl.subplot(gs[3,0], sharex=ax)
        ax.errorbar(lc_data.date, lc_data.flux, yerr=lc_data.flux_err, ls='', marker='.', color='b')
        xx = np.linspace(lc_data.date.min(), lc_data.date.max(), 100)
        xx.sort()
        try:
            tmax = f.pars['tmax'].full
        except:
            tmax = np.zeros(N)
        xdata = salt3.plotting.plots.get_xdata_for_model_evaluation2(N, tmax, lc_data.band_set, npoints=200)
        v = model(f.pars.free, data=xdata, jac=False)
        #4(6)
        for sn in lc_data.sn_set:
            for bd in model.bands_unique:
                idx = (xdata.sn == sn) & (xdata.band == bd)
                ax.plot(xdata.date[idx], v[idx], ls='-', color=pl.cm.jet(sn/xdata.sn.max()))

        dif.errorbar(lc_data.date, f(x,jac=False)*lc_data.flux_err, yerr=lc_data.flux_err, marker='.', ls='', color='b')
        dif.grid(1)
        dif.set_xlabel('x')
        dif.set_ylabel('res')
        pl.subplots_adjust(hspace=0.05)

        for bd in model.bands_unique:
            pl.figure()
            xx = np.linspace(-50., 50., 10000)
            jacobian = model.basis.eval(xx)
            pl.plot(xx, jacobian @ model.pars[f"theta_{bd.decode('UTF-8')}"].full, 'r-')
            pl.xlabel('x')
            pl.ylabel('LC template (spline model only)')
            pl.title(bd.decode('UTF-8'))
            pl.grid(1)

    return lc_data, f, cons, reg, x






def test_model_derivatives(N=200, bands = bands, seed=0):
    """                                                                                                 
    """
    lc_data = salt3.simulations.generator.generate_lc_multiband(N=N, bands=bands,
                                                    tmax_range=(-70, 70.),
                                                    sig_range=(-1, 1.),
                                                    norm_range=(1000., 1000.),
                                                    seed=seed) 
    
    model = salt3.models.models.MultiBandsModel(lc_data, rng_spline = [-100, 100], grid_size=250)
    f = salt3.models.models.VarModelResiduals(lc_data, model)

    jacobian, JJ = check_grad(model, model.pars.free)
    D = np.abs(jacobian-JJ)
    np.testing.assert_allclose(np.zeros_like(D), D, atol=1.E-8)

    jacobian, JJ = check_grad(f, model.pars.free)
    D = np.abs(jacobian-JJ)
    np.testing.assert_allclose(np.zeros_like(D), D, atol=1.E-6)

    model.pars.free[:] += np.random.uniform(-1., 1., len(model.pars.free))
    jacobian, JJ = check_grad(model, model.pars.free)
    D = np.abs(jacobian-JJ)
    np.testing.assert_allclose(np.zeros_like(D), D, atol=1.E-6)

    jacobian,JJ = check_grad(f, model.pars.free)
    D = np.abs(jacobian-JJ)
    np.testing.assert_allclose(np.zeros_like(D), D, atol=1.E-6)
    
    ###  passe pas
    # model.pars.full[:] += np.random.uniform(-1., 1., len(model.pars.full))                          
    # jacobian, JJ = check_grad(model, model.pars.free)
    # D = np.abs(jacobian-JJ)
    # np.testing.assert_allclose(np.zeros_like(D), D, atol=1.E-6)                                     

    # jacobian,JJ = check_grad(f, model.pars.free)
    # D = np.abs(jacobian-JJ)
    # np.testing.assert_allclose(np.zeros_like(D), D, atol=1.E-6) 

    # Variance model derivative test :

    variance_model = salt3.models.variancemodels.VarianceMultibandModel(model, var_init = 0.01)
    f = salt3.models.models.VarModelResiduals(lc_data, model, variance_model = variance_model)

    D0, D1 = check_grad_var(f, model.pars.free, variance_model.pars.free)
    D = np.abs(D0-D1)
    np.testing.assert_allclose(np.zeros_like(D), D, atol=1.E-6)

    model = salt3.models.models.SiftoModel(lc_data, grid_size = 250, rng_spline = [-100, 100])
    variance_model = salt3.models.variancemodels.VarianceMultibandModel(model, var_init = 0.01)
    f = salt3.models.models.VarModelResiduals(lc_data, model, variance_model = variance_model)

    D0, D1 = check_grad_var(f, model.pars.free, variance_model.pars.free)
    D = np.abs(D0-D1)
    np.testing.assert_allclose(np.zeros_like(D), D, atol=1.E-6)
    
def test_1D_multibandmodelfullfit_nr():
    x_target = np.load(locate_test_datafile('test_1D_multibandmodel_solution.npy'))
    lc, f, cons, reg, x = fit_1dsplinemodelmultiband_same_lcs(N = 100, mu_regularization=10, seed = 0)
    np.testing.assert_allclose(x_target, x, atol=1.E-6)
    np.testing.assert_allclose(np.zeros_like(cons(x, jac=False)), cons(x, jac=False), atol=1.E-8)



def fit_1dmultibandmodel(N=100, bands = bands,
                         minimize=minimize_nr_var_quad, mu_regularization=20.,
                         plot=False, seed=None):
    """
    given                                                                                               
    """
    logging.info('generating data...')
    lc_data = salt3.simulations.generator.generate_lc_multiband(N=N, bands=bands,
                                                    tmax_range=(-50., 50.),
                                                    sig_range=(-0.5, 0.5),
                                                    norm_range=(100., 10000.),
                                                    seed=seed)

    model = salt3.models.models.MultiBandsModel(lc_data, grid_size=100)
    # fit initialization
    logging.info('initialization...')
    model.pars.full[:] = salt3.models.models.init_splinemodel1dmultiband(model).full[:]
    p_init = model.pars.copy()

    # minimization
    logging.info('minimization...')
    f = salt3.models.models.VarModelResiduals(lc_data, model)
    cons = salt3.models.constraints.NonIntegralConstraints_4MultibandModel(f.pars, model,
                                                               band_tmax='B')
    reg = salt3.models.regularizations.SplineParamsRegularizationMultiband(f.pars,
                                                           bands = model.bands_unique,
                                                           mu=mu_regularization)
    x = minimize(f, f.pars.free, cons=cons, reg=reg)

    if plot:
        import pylab as pl
        import matplotlib as mpl
        salt3.plotting.plots.plot_all_lcs_stacked_multiband_stretch(lc_data, model, sifto_like = False)
        
        # plot the template itself
        for bd in model.bands_unique:
            pl.figure()
            xx = np.linspace(-50., 50., 10000)
            jacobian = model.basis.eval(xx)
            pl.plot(xx, jacobian @ model.pars[f"theta_{bd.decode('UTF-8')}"].full, 'r-')
            pl.xlabel('x')
            pl.ylabel('LC template (spline model only)')
            pl.grid(1)
            pl.title('spline model band : ' + bd.decode('UTF-8'))

        # and the residuals
        pl.figure()
        pl.hist(f(f.pars.free, jac=False), bins=100)
        pl.xlabel('residuals')

    return lc_data, f, model, cons, reg, p_init, x


def test_1dmultibandmodelfullfit_with_init_200():
    ff = np.load(locate_test_datafile('test_1D_multibandmodel_fullfit_with_init_solution.npz'))
    x_target = ff['x_200']
    lc, f, model, cons, reg, p_init, x = fit_1dmultibandmodel(N=200, minimize=minimize_nr_var_quad, mu_regularization=100., plot=False, seed=0)
    x = x[0]
    np.testing.assert_allclose(x_target, x, atol=1.E-8)
    c = cons(x, jac=False)
    np.testing.assert_allclose(np.zeros_like(c), c, atol=1.E-8)

def test_1dmultibandmodelfullfit_with_init_500():
    ff = np.load(locate_test_datafile('test_1D_multibandmodel_fullfit_with_init_solution.npz'))
    x_target = ff['x_500']
    lc, f, model, cons, reg, p_init, x = fit_1dmultibandmodel(N=500, minimize=minimize_nr_var_quad, mu_regularization=200., plot=False, seed=0)
    x = x[0]
    np.testing.assert_allclose(x_target, x, atol=1.E-8)
    c = cons(x, jac=False)
    np.testing.assert_allclose(np.zeros_like(c), c, atol=1.E-8)

def test_1dmultibandmodelfullfit_with_init_1000():
    ff = np.load(locate_test_datafile('test_1D_multibandmodel_fullfit_with_init_solution.npz'))
    x_target = ff['x_1000']
    lc, f, model, cons, reg, p_init, x = fit_1dmultibandmodel(N=1000, minimize=minimize_nr_var_quad, mu_regularization=400., plot=False, seed=0)
    x = x[0]
    np.testing.assert_allclose(x_target, x, atol=1.E-8)
    c = cons(x, jac=False)
    np.testing.assert_allclose(np.zeros_like(c), c, atol=1.E-8)


#@profile
def fit_multibandsmodel_var(N=200, minimize=minimize_nr_var_quad,
                            mu_regularization=10., plot=False, seed=None, sifto_like = True, order = 4, timing = False):
    """
    """
    lc_data = salt3.simulations.generator.generate_lc_multiband(N=N, bands=bands,
                                                    tmax_range=(-50., 50.),
                                                    sig_range=(-0.5, 0.5),
                                                    norm_range=(100., 10000.),
                                                    seed=seed)
    if sifto_like:
        model = salt3.models.models.SiftoModel(lc_data, grid_size = 250, 
                                        rng_spline = [-100, 100])
    else :
        model = salt3.models.models.MultiBandsModel(lc_data, grid_size=250,
                                             rng_spline = [-100, 100])
        
    variance_model = salt3.models.variancemodels.VarianceMultibandModel(model, var_init = 0.01)
    f = salt3.models.models.VarModelResiduals(lc_data, model, variance_model = variance_model)
    cons = salt3.models.constraints.NonIntegralConstraints_4MultibandModel(f.model.pars, model,
                                                               band_tmax='B',
                                                               sifto_like = sifto_like)
    reg = salt3.models.regularizations.SplineParamsRegularizationMultiband(f.model.pars,
                                                           bands = model.bands_unique,
                                                           mu=mu_regularization,
                                                           deriv = False)
    # initialization fit (spline parameters only)
    model.pars.full[:] = salt3.models.models.init_splinemodel1dmultiband(model, mu_regularization =.01,
                                                                  deriv = False,
                                                                  sifto_like = sifto_like).full[:]
    if plot :
        salt3.plots.plot_all_lcs_stacked_multiband_stretch(lc_data, model, sifto_like = sifto_like)
    if timing : 
        t0 = time.time()
    x, log0  = minimize(f, f.model.pars.free, g0 = None, reg=reg, cons=cons, log = [])
    
    if plot:
        salt3.plots.plot_all_lcs_stacked_multiband_stretch(lc_data, model,
                                                           sifto_like = sifto_like,
                                                           varmodel = variance_model)
    if timing : 
        t1 = time.time()

    x, g, log = minimize(f, f.model.pars.free, g0 = f.VarianceModel.pars.free,
                         reg=reg, cons=cons, log = [])
    if timing : 
        t2 = time.time()

    if plot :
        salt3.plotting.plots.plot_all_lcs_stacked_multiband_stretch(lc_data, model,
                                                           sifto_like = sifto_like,
                                                           varmodel = variance_model)
        
        pl.figure()
        pl.hist(np.abs(g), bins = 'auto')
        pl.vlines(0.01, 0, 10, color = 'b',label = 'initializarion')
        pl.vlines(0.02, 0, 10, color = 'r', label = 'implementation')
        pl.text(np.abs(g).mean(), 10, f'm : {np.abs(g).mean() : .3f}, \n s : {np.abs(g).std() : .3f}')
        pl.grid(1)
        pl.legend()
            
    if timing : 
        return lc_data, f, cons, reg, x, g, log0, log, t0, t1, t2
    return lc_data, f, cons, reg, x, g



def timing(N = [10, 50, 100, 500, 1000, 5000, 10000], stock = True):

    if stock :
        T = np.array([0.024266421794891357, 0.0399980147679646, 0.0516820028424263, 0.24664335963369785, 0.4919382160545414, 2.039620816707611, 4.158022114208767  ])
        T_err = np.array([0.04793691635131836, 0.05689263343811035, 0.10866343975067139, 0.45550668239593506, 0.9045883019765218, 5.144795179367065, 11.023170948028564 ])
        ite = np.array([4, 6, 32, 87, 234, 16, 14])
        ite_err = np.array([2, 2, 2, 2, 3, 2, 2])
    else : 
        T = []
        T_err = []
        ite = []
        ite_err = []
        
        for n in N:
            lc_data, f, cons, reg, x, g, log0, log, t0, t1, t2 = fit_multibandsmodel_var(N=n, mu_regularization = n/100, timing = True)
        
            T.append((t1 - t0)/(log0[-1][0]+1))
            T_err.append((t2 - t1)/(log[-1][0]+1))
            ite.append(log0[-1][0]+1)
            ite_err.append(log[-1][0]+1)
            print('\n', n, np.array(T)[-1], np.array(T_err)[-1])
        
        T = np.array(T)
        T_err = np.array(T_err)
    x = np.array(N)
    fig = pl.figure()
    ax = fig.add_subplot(1, 1, 1)

    ax.plot(x, T, color = 'r', ls = '',
            marker = '.', label ='Fit flux model', markersize=20)
    ax.plot(x, T_err, color = 'b', ls = '',
            marker = '^', label ='With error model', markersize=20) 
    ax.set_xscale('log')
    ax.set_yscale('log')
    ax.grid(which='both')
    parameters = {'axes.labelsize': 20,
                  #'axes.titlesize': 300,
                  'xtick.labelsize': 25,
                  'ytick.labelsize': 25}
    pl.rcParams.update(parameters)
    ax.grid(which='minor', alpha=0.5, linestyle='-')
    ax.set_ylabel('time per iteration (in sec)')
    ax.set_xlabel('Number of SNe (5 bands)')
    pl.legend(fontsize=20)



    
def test_1dmultiband_sifto_model():
    ff = np.load(locate_test_datafile('test_1D_multiband&Sifto_model_var_with_init_solution.npz'))
    x_sifto_target = ff['x_sifto']
    x_multi_target = ff['x_multi']
    g_sifto_target = ff['g_sifto']
    g_multi_target = ff['g_multi']  
    lc_data, f, cons_m, reg, x_multi, g_multi = fit_multibandsmodel_var(mu_regularization=50.,
                                                                        plot = False,
                                                                        sifto_like = False,
                                                                        seed = 0)

    lc_data, f, cons_s, reg, x_sifto, g_sifto = fit_multibandsmodel_var(mu_regularization=50.,
                                                                        plot = False,
                                                                        sifto_like = True,
                                                                        seed = 0)
    np.testing.assert_allclose(x_sifto_target, x_sifto, atol=1.E-8)
    np.testing.assert_allclose(x_multi_target, x_multi, atol=1.E-8)
    np.testing.assert_allclose(g_sifto_target, g_sifto, atol=1.E-8)
    np.testing.assert_allclose(g_multi_target, g_multi, atol=1.E-8)

    cs = cons_s(x_sifto, jac=False)
    cm = cons_m(x_multi, jac=False)
    np.testing.assert_allclose(np.zeros_like(cm), cm, atol=1.E-8)
    np.testing.assert_allclose(np.zeros_like(cs), cs, atol=1.E-8)

# if __name__ == "__main__":
#      test_model_derivatives(N=200)
#      test_1dmultibandmodelfullfit_with_init_200()
#      test_1dmultibandmodelfullfit_with_init_500()
#      test_1dmultibandmodelfullfit_with_init_1000()
#      test_1dmultiband_sifto_model()



import numpy as np
import scipy.sparse as sparse
import pandas
from matplotlib import pyplot as pl

from nacl.loglikelihood import LogLikelihood
from nacl.minimize import Minimizer
from nacl.models.toy2d import spline
from scipy.sparse import coo_matrix, dia_matrix, csc_matrix, dok_matrix
from sksparse import cholmod


def main_no_adapted_regularization_2D(n0=50, n1=50, b0_size=25, b1_size=25):
    """
    """
    gen = spline.ToyDataGenerator(n0=n0, n1=n1, b0_size=b0_size, b1_size=b1_size, error_pedestal=0., delta=5.)
    d = gen.generate()
    reg = spline.Regularization(block_name='theta', mu0=1., mu2=1.)

    model = spline.Model(gen.x0, gen.x1, gen.basis.bx.grid, gen.basis.by.grid, order=4)
    ll = LogLikelihood(model, reg=[reg], data=gen.get_data())
    #ll = LogLikelihood(model, reg=None, data=gen.get_data())
    v, grad, hess = ll(ll.pars.free, deriv=1)

    minz = Minimizer(ll)
    p_init = ll.pars.copy()
    p_init['theta'].full[:] = np.random.rand(len(p_init['theta'].full))
    minz.minimize_lm(p_init=p_init.free, lamb=1.E-6)
    spline.plot_model_and_data(minz.log_likelihood.model, minz.log_likelihood.pars, minz.log_likelihood.data)

def hand_made_fit(n0=50, n1=50, b0_size=25, b1_size=25):
    """
    Adding this to see if I can get insight on where the fitting problem is
    """
    gen = spline.ToyDataGenerator(n0=n0, n1=n1, b0_size=b0_size, b1_size=b1_size, error_pedestal=0., delta=5.)
    d = gen.generate()
    data = gen.get_data()

    def fit(x0, x1, basis, y, yerr):
        """a fit of the data -- the simplest possible
        """
        J = basis.eval(np.array(x0), np.array(x1))
        w = 1. / yerr**2
        W = dia_matrix((w, [0]), shape=(len(y), len(y)))
        H = J.T @ W @ J
        B = J.T @ W @ y
        fact = cholmod.cholesky(H, beta=1.E-6)
        theta = fact(B)

        resid = y-J @ theta
        return resid, theta
        
    r, t = fit(data.xx0, data.xx1, gen.basis, data.y, data.yerr)
    
    return r
    
    



import numpy as np
import pylab as pl
import logging
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                    level=logging.INFO)
from nacl.dataset import TrainingDataset
from nacl import minimize
from nacl.fit import TrainSALT2Like
from nacl.models import salt2
try:
    from nacl.models.salt2 import wl_extend
except:
    logging.info(f'wl_extend not imported')
import pandas as pd
import os
os.environ['KMP_DUPLICATE_LIB_OK']='True'
pl.ion()

sne_full = pd.read_hdf('../sandbox/snf_tds_not_clean.hd5', key='sn_data')
spec_full = pd.read_hdf('../sandbox/snf_tds_not_clean.hd5', key='spectrophotometric_data')
#sne_full = pd.read_hdf('../sandbox/snf_tds_spline.hd5', key='sn_data')
#spec_full = pd.read_hdf('../sandbox/snf_tds_spline.hd5', key='spectrophotometric_data')
spec_full = spec_full.assign(i_basis = [0 for x in range(len(spec_full))])
tds_full = TrainingDataset(sne_full, spectrophotometric_data=spec_full)

def filter_tds(tds, min_wavelength=3600.):
    idx = tds.spectrophotometric_data.wavelength <= min_wavelength
    logging.info(f'invalidating {idx.sum()} measurements outside wl range')
    tds.spectrophotometric_data.valid[idx] = 0
    tds.kill_sne(['PTF12iiq', 'LSQ13aiz', 'PTF11qzq', 'SN2006X'])
    # tds.kill_sne(['PTF12ikt', 'PTF10zdk', 'PTF11bgv'])
    # tds.kill_sne(['SN2004gc', 'PTF11kly', 'SN2005cg'])
    tds.kill_photometric_spectra([1121, 3116, 3307, 1803, 690,
                                   788, 1670, 1056, 1057, 732, 1614,
                                   817, 1224, 1225, 1226, 1227, 1228,
                                  1229, 1230, 1231, 1232, 1233, 1234,
                                  1235, 1236, 1237, 1238, 1239, 1240,
                                  1241, 1242, 1243, 1244, 1245, 1246,
                                  1580, 1854])

    flux = tds.spectrophotometric_data.flux
    fluxerr = tds.spectrophotometric_data.fluxerr
    ey = np.sqrt(fluxerr**2 + 0.05**2 * flux**2)
    tds.spectrophotometric_data.nt['fluxerr'] = ey

    tds.compress()

    return tds

def plot_chi2_spec(tds, v):
    """
    """
    # residuals
    pl.figure()
    x = np.arange(len(v))
    idx = tds.spectrophotometric_data.valid > 0
    res = tds.get_all_fluxes() - v
    yerr=tds.get_all_fluxerr()
    wres = res / yerr
    pl.figure()
    spec_chi2 = np.bincount(np.int64(tds.spectrophotometric_data.spec[idx]),
                            wres[idx]**2)
    spec_nmeas = np.bincount(np.int64(tds.spectrophotometric_data.spec[idx]))
    rchi2_spec = spec_chi2 / spec_nmeas
    pl.plot(rchi2_spec, 'ro')
    pl.title('spectrum $\chi^2$')
    pl.xlabel('spec_index')
    pl.ylabel('$\chi^2$')
    
    pl.figure()
    pl.hist(rchi2_spec, bins=80)
    pl.title('spec $\chi^2$/n_meas')
    pl.ylabel('Counts')
    pl.xlabel('spec $\chi^2$/n_meas')
    
def remove_chi2_spec(tds, model, pars, thresh=1e+4):
    """
    """
    v = model(pars, jac=False)
    idx = tds.spectrophotometric_data.valid > 0
    res = tds.get_all_fluxes() - v
    yerr=tds.get_all_fluxerr()
    wres = res / yerr
    pl.figure()
    spec_chi2 = np.bincount(np.int64(tds.spectrophotometric_data.spec[idx]),
                            wres[idx]**2)
    spec_nmeas = np.bincount(np.int64(tds.spectrophotometric_data.spec[idx]))
    rchi2_spec = spec_chi2 / spec_nmeas
    
    spec_to_kill = np.where(rchi2_spec > thresh)[0]
    tds.kill_photometric_spectra(spec_to_kill)
    tds.compress()
    tds.spectrophotometric_data.valid = np.int64(tds.spectrophotometric_data.valid)
    

def plot_spectrum(tds, v, spec, var):
    """
    Plotting fitted spectra
    """
    pl.figure()
    idx = tds.spectrophotometric_data.spec == spec
    pl.plot(tds.spectrophotometric_data.wavelength[idx],
            tds.spectrophotometric_data.flux[idx], 'r+')
    pl.errorbar(tds.spectrophotometric_data.wavelength[idx],
                tds.spectrophotometric_data.flux[idx],
                yerr=tds.spectrophotometric_data.fluxerr[idx],
                ls='', marker='.', label='data')
    pl.plot(tds.spectrophotometric_data.wavelength[idx], v[idx],
            color='r', lw=2, ls='-', zorder=10,
            label='trained model')
    pl.fill_between(tds.spectrophotometric_data.wavelength[idx], 
                    v[idx]-np.sqrt(var[idx]), v[idx]+np.sqrt(var[idx]), alpha = 0.5)
    pl.title(f'spectrum #{spec}', fontsize=16)
    pl.xlabel('observer frame $\lambda\ [\AA]$', fontsize=16)
    pl.ylabel('coefficient', fontsize=16)
    pl.legend(loc='best')
    
def plot_spectrum_test(tds, v, spec=0, var=None):
    """
    Plotting fitted spectra
    """
    basis = tds.basis
    ii = tds_full.spectrophotometric_data.spec == spec
    wl = tds_full.spectrophotometric_data.wavelength[ii] / (1.+tds_full.spectrophotometric_data.z[ii])
    flux_true = tds_full.spectrophotometric_data.flux[ii]
    fluxerr_true = tds_full.spectrophotometric_data.fluxerr[ii]
    #w = 1/fluxerr_true
    #W = sparse.dia_matrix((w**2, 0), shape=(len(wl),len(wl)))
    
    J = basis.eval(np.array(wl))
    J = J.tocsr()
    idx = tds.spectrophotometric_data.spec == spec
    i_basis = tds.spectrophotometric_data.i_basis[idx]
    flux = tds.spectrophotometric_data.flux[idx]
    
    JJ = J[:,i_basis]
    
    y = JJ @ flux
    y_fit = JJ @ v[idx]
    y_err_model = JJ @ var[idx]    
    #pl.plot(wl, y)
    #pl.plot(wl, flux_true)
    
    
    pl.figure()
    pl.plot(wl,
            flux_true, 'r+')
    pl.errorbar(wl,
                flux_true,
                yerr=fluxerr_true,
                ls='', marker='.', label='data')
    pl.plot(wl, y_fit,
            color='r', lw=2, ls='-', zorder=10,
            label='trained model')
    pl.fill_between(wl, 
                    y_fit-np.sqrt(y_err_model), y_fit+np.sqrt(y_err_model), alpha = 0.5)
    pl.title(f'spectrum #{spec}', fontsize=16)
    pl.xlabel('restframe frame $\lambda\ [\AA]$', fontsize=16)
    pl.ylabel('flux', fontsize=16)
    pl.legend(loc='best')

def plot_stacked_residuals_test(tds, model, pars, var, v=None):
    """
    Pulls with snake error
    """
    if v is None:
        v = model(pars, jac=False)
    basis = tds.basis
    y_fit = []
    y_err = []
    y_true = []
    yerr_true = []
    wave = []
    phase = []
    for spec in range(int(np.max(tds_full.spectrophotometric_data.spec_index))):
        ii = tds_full.spectrophotometric_data.spec_index == spec
        wl = tds_full.spectrophotometric_data.wavelength[ii] / (1.+tds_full.spectrophotometric_data.z[ii])
        ph = (tds_full.spectrophotometric_data.mjd[ii] - pars['tmax'].full[tds_full.spectrophotometric_data.sn_index][ii]) / (1 + tds_full.spectrophotometric_data.z[ii])
        wave.append(wl)
        phase.append(ph)
        flux_true = tds_full.spectrophotometric_data.flux[ii]
        y_true.append(flux_true)
        fluxerr_true = tds_full.spectrophotometric_data.fluxerr[ii]
        yerr_true.append(fluxerr_true)
        
        J = basis.eval(np.array(wl))
        J = J.tocsr()
        idx = tds.spectrophotometric_data.spec_index == spec
        i_basis = tds.spectrophotometric_data.i_basis[idx]
        JJ = J[:,i_basis]
        if len(JJ @ v[idx]) == 0:
            print('no data : spec ', spec)
        y_fit.append(JJ @ v[idx])
        y_err.append(JJ @ var[idx])
    y_fit = np.hstack(y_fit)
    y_err = np.hstack(y_err)
    y_true = np.hstack(y_true)
    yerr_true = np.hstack(yerr_true)
    print(len(y_fit))
    wl = np.hstack(wave)
    ph = np.hstack(phase)

    pl.figure(figsize=(14, 10))
    
    res = y_true - y_fit
    wres = res / (yerr_true + np.sqrt(y_err) )
    
    ii = ph < 50.
    if len(ph[ii]) != len(ph):
        logging.warning(f'Phase > 50 exists fit not very good')
    
    pl.hexbin(wl[ii], ph[ii], wres[ii], vmin=-2.5, vmax=2.5)
    pl.title('pulls with simple snake', fontsize=16)
    pl.xlabel('restframe $\lambda [\AA]$', fontsize=16)
    pl.ylabel('restframe phase [days]', fontsize=16)
    pl.colorbar()
  
def plot_stacked_residuals(tds, model, pars, var, v=None):
    """
    Pulls with snake error
    """
    if v is None:
        v = model(pars, jac=False)

    wl = tds.spectrophotometric_data.wavelength / (1. + tds.spectrophotometric_data.z)
    ph = (tds.spectrophotometric_data.mjd - pars['tmax'].full[tds.spectrophotometric_data.sn_index]) / (1 + tds.spectrophotometric_data.z)

    pl.figure(figsize=(14, 10))
    idx = tds.get_valid() > 0
    res = tds.get_all_fluxes() - v
    wres = res / (tds.get_all_fluxerr() + np.sqrt(var))
    
    ii = ph[idx] < 50.
    if len(ph[idx][ii]) != len(ph[idx]):
        logging.warning(f'Phase > 50 exists fit not very good')
    
    pl.hexbin(wl[idx][ii], ph[idx][ii], wres[idx][ii], vmin=-2.5, vmax=2.5)
    pl.title('pulls with simple snake', fontsize=16)
    pl.xlabel('restframe $\lambda [\AA]$', fontsize=16)
    pl.ylabel('restframe phase [days]', fontsize=16)
    pl.colorbar()

def plot_chi2(tds, model, pars, var, v=None):
    """
    Chi2 with snake error
    """
    if v is None:
        v = model(pars, jac=False)
    wl = tds.spectrophotometric_data.wavelength / (1. + tds.spectrophotometric_data.z)
    ph = (tds.spectrophotometric_data.mjd - pars['tmax'].full[tds.spectrophotometric_data.sn_index]) / (1 + tds.spectrophotometric_data.z)
    ii = ph < 50
    pl.figure(figsize=(14, 10))
    idx = tds.get_valid() > 0
    res = tds.get_all_fluxes() - v
    wres = res / (tds.get_all_fluxerr() + np.sqrt(var))
    chi2, _, _ = np.histogram2d(wl[ii], ph[ii], bins=(400,50), weights=wres[ii]**2)
    cc, _, _ = np.histogram2d(wl[ii], ph[ii], bins=(400,50))
    rchi2 = chi2 / cc
    pl.imshow(rchi2.T[::-1,:], vmin=0, vmax=10,
              aspect='auto',
              interpolation='none',
              extent=[wl[ii].min(), wl[ii].max(), ph[ii].min(), ph[ii].max()])
    pl.title('local $\chi^2$', fontsize=16)
    pl.xlabel('restframe $\lambda [\AA]$', fontsize=16)
    pl.ylabel('restframe phase [days]', fontsize=16)
    pl.colorbar()

if __name__ == '__main__':
    tds = TrainingDataset.read_hdf('../sandbox/snf_projected_200.hd5')
    #tds = TrainingDataset.read_hdf('../sandbox/snf_projected_450.hd5')
    #tds = TrainingDataset.read_hdf('../sandbox/snf_projected_test.hd5')
    
    tds = filter_tds(tds, min_wavelength=3600.)
    
    tds.spectrophotometric_data.valid = np.int64(tds.spectrophotometric_data.valid)  #should be looked at
    
    try:
        grid = wl_extend()
    except:
        gird = np.linspace(2000., 11000., 200)
    
    #model = salt2.SALT2Like(tds, wl_grid = grid)
    #pars = model.init_pars()
    #remove_chi2_spec(tds, model, pars, thresh=1e+2)
    
    model = salt2.SALT2Like(tds, wl_grid = grid)
    pars = model.init_pars()
    v_init = model(pars)
    tm = TrainSALT2Like(tds, mu_reg=1, mu_cons=1e-10)
    #res = tm.fit(block_names=['X0', 'X1', 'col', 'tmax'], pars=pars, fit_error_model=True)
    #res = tm.fit(block_names=['X0', 'X1', 'col', 'tmax', 'M0', 'M1', 'CL'], pars=pars, fit_error_model=False)
    res = tm.fit_by_blocks(block_names=['X0', 'X1', 'col', 'tmax'], pars=pars, fit_error_model=True)
    minz = res['minz']
    m = minz.log_likelihood.model
    v = m(res['pars'])
    try:
        v_err = minz.log_likelihood.variance_model(res['pars'])
    except:
        v_err = np.zeros(len(v))
    
    plot_spectrum_test(tds, v, 14, v_err)
    #plot_spectrum_test(tds, v_init, 0, v_err)
    plot_spectrum(tds, v, 14, v_err)
    #plot_spectrum(tds, v_init, 0, v_err)
    plot_spectrum_test(tds, v, 24, v_err)
    plot_spectrum(tds, v, 24, v_err)
    plot_stacked_residuals(tds, m, res['pars'], v_err, v=v)
    #plot_stacked_residuals_test(tds, m, res['pars'], v_err, v=v)
    #plot_stacked_residuals(tds, model, pars, v_err, v=v_init)
    #import test_snf
    #test_snf.compare_pars(tds, res['pars'])    
    #plot_chi2(tds, m, res['pars'], v_err, v=v)



#!/usr/bin/env python3

import logging
from argparse import ArgumentParser
from pathlib import Path
import glob

import numpy as np
import pandas

from nacl.dataset import TrainingDataset
from nacl.specutils import Spec, clean_and_project_spectra
from nacl.models import salt2
from lemaitre import bandpasses

bandpass_convert = {
    'MEGACAMPSF::g': 'megacam6::g',
    'MEGACAMPSF::r': 'megacam6::r',
    'MEGACAMPSF::i2': 'megacam6::i2',
    'MEGACAMPSF::z': 'megacam6::z',
    'MEGACAMPSF::i': 'MEGACAMPSF::i',

    'ztfg': 'ztf::g',
    'ztfr': 'ztf::r',
    'ztfi': 'ztf::I',

    'ztf::g': 'ztf::g',
    'ztf::r': 'ztf::r',
    'ztf::I': 'ztf::I',

    'megacam6::g': 'megacam6::g',
    'megacam6::r': 'megacam6::r',
    'megacam6::i2': 'megacam6::i2',
    'megacam6::z': 'megacam6::z',
    'megacam6::i': 'MEGACAMPSF::i',
}


def load_sample(sample='SNLS', init_sn_loc=True, input_dir='./'):
    """
    load and sanitize the SNLS data
    """
    input_dir = Path(input_dir)

    data = []
    for nm  in ['data', 'lc', 'sp']:
        path = glob.glob(str(input_dir.joinpath(f'{nm}_{sample}*parquet*')))
        assert len(path) == 1
        data.append(pandas.read_parquet(path[0]))
    sne, lc, sp = data
    # lc = pandas.read_csv(input_dir.joinpath(f'lc_{sample}.parquet.gzip'))
    # sp = pandas.read_csv(input_dir.joinpath(f'sp_{sample}.parquet.gzip'))

    # purge all the SN which could not be fitted with sncosmo
    idx = sne['x0'] != 999.9
    logging.info(f'removed {(~idx).sum()} SNe which could not be fitted with sncosmo')
    sne = sne[idx]

    # drop the index and Unnamed fields
    sne = sne
    
    # purge all the LC data taken with bandpass megacampsf::i
    idx = (lc.band == 'MEGACAMPSF::i') | (lc.band == 'megacampsf::i') | (lc.band == 'megacam6::i')
    lc = lc[~idx]
    logging.info(f'removed {(idx).sum()} LC data points taken in band megacampsf::i')
    
    # purge all the LC data with no counterpart in the SN index
    idx = lc.name.isin(sne.name)
    lc = lc[idx]
    logging.info(f'removed {(~idx).sum()} LC data points with no entry in the SN index')
    lc.sn = lc.name

    # purge all the spectra with no counterpart in the SN index
    idx = sp.name.isin(sne.name)
    sp = sp[idx]
    logging.info(f'removed {(~idx).sum()} SP data points with no entry in the SN index ')
    sp.sn = sp.name
    
    # correct the valid field
    sp.loc[sp.valid.isna(), 'valid'] = 1
    sp['valid'] = sp.valid.astype(int)
    
    # we use names to identify the SNe
    sne.sn = sne.name
    lc.sn = lc.name
    sp.sn = sp.name

    # update the bandpass names
    lc['band'] = [bandpass_convert[b] for b in lc.band]
    if 'survey' not in sne.columns:
        lc['survey'] = sample

    if init_sn_loc:
        lc['x'] = 0.
        lc['y'] = 0.
        lc['ccd'] = 0
        lc['amp'] = 0
        lc['sensor_id'] = 0

    sp['i_basis'] = -1
    
    return sne, lc, sp



if __name__ == '__main__':
    parser = ArgumentParser(description='create a TrainingDataset from the PETS files')
    parser.add_argument('-o', '--output',
                        default='lemaitre',
                        help='output file name')
    parser.add_argument('--dump-uncompressed-tds',
                        action='store_true',
                        help='whether to dump the uncompressed training dataset')
    parser.add_argument('--format',
                        default='parquet',
                        help='output format: parquet or hdf5')
    parser.add_argument('-C', '--compress',
                        action='store_true',
                        help='project spectra on the model basis')
    parser.add_argument('--samples', nargs='*',
                        default=['ztf', 'SNLS'],
                        help='samples to combine')
    parser.add_argument('input_dir',
                        help='where to find the PETS results')
    args = parser.parse_args()
    
    
    #    parser.add_argument('--basis', default='',
    #                        help='project spectra on the model basis')
    
    #    logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
    #                        level=logging.INFO)

    logging.info('loading the default filter lib')
    fl = bandpasses.get_filterlib()

    sne, lc, sp = [], [], []
    lc_offset = 0
    spec_offset = 0
    for sample_name in args.samples:
        logging.info(f'loading {sample_name}')
        sn_data, lc_data, sp_data = load_sample(sample_name, input_dir=args.input_dir)
        lc_data['lc'] += lc_offset
        sp_data['spec'] += spec_offset
        #        lc['lc'] = lc['lc'] + lc_offset
        #        sp['spec'] = sp['spec'] + spec_offset
        lc_offset = lc_data['lc'].max() + 1
        spec_offset = sp_data['spec'].max() + 1
        
        sne.append(sn_data)
        lc.append(lc_data)
        sp.append(sp_data)

    logging.info('building the uncompressed Training Dataset')        
    tds = TrainingDataset(pandas.concat(sne),
                          pandas.concat(lc),
                          pandas.concat(sp),
                          filterlib=fl)
    tds.compress()
    
    if args.output and args.dump_uncompressed_tds:
        p = Path(args.output)
        if 'parquet' in args.format:
            tds.to_parquet(p.with_suffix('.uncompressed.parquet'))
        elif 'hd5' in args.format or 'hdf' in args.format:
            tds.to_hdf(p.with_suffix('.uncompressed.hd5'))

    logging.info('default model')
    model = salt2.SALT2Like(tds)
    
    logging.info('compression...')
    projected_spectra, in_error = clean_and_project_spectra(tds, model.basis.bx)
    logging.info('done: in_error: {len(in_error)}')

    logging.info(f' -> {args.output}')
    ptds = TrainingDataset(tds.sn_data.nt, lc_data=tds.lc_data.nt,
                           spec_data=np.rec.array(np.hstack(projected_spectra)),
                           basis=model.basis.bx,
                           filterlib=fl)
    
    if args.output:
        p = Path(args.output)
        if 'parquet' in args.format:
            ptds.to_parquet(p.with_suffix('.compressed.parquet'))
        elif 'hd5' in args.format or 'hdf' in args.format:
            ptds.to_hdf(p.with_suffix('.compressed.hd5'))
    



import numpy as np
import scipy.sparse
import croaks
from saunerie.fitparameters import FitParameters
from salt3.minimizers import minimize_nr_var_quad, minimize_nr, minimize
import salt3.models

def gen_sincos(N=100, A=1., B=-0.2, omega_a=0.72, omega_b=0.52, ey=0.1,
               x=None, y=None, yerr=None, plot=False,
               seed=None):
    """
    Generate a dataset
    """

    if isinstance(seed, int):
        np.random.seed(seed)
        
    if x is None or y is None or yerr is None:
        x = np.random.uniform(-5., 5., N)
        y = A*np.sin(omega_a*x) + B*np.cos(omega_b*x)
        y += np.random.normal(scale=0.1, size=N)
        yerr = np.full(N, ey)

    if plot:
        import pylab as pl
        pl.figure()
        pl.errorbar(x, y, yerr=ey, ls='', color='b', marker='.')

    data = np.rec.fromarrays((x,y,yerr), names=['x', 'y', 'ey'])
    return croaks.DataProxy(data, wavelength='x', flux='y', flux_err='ey')


class SinCosToyModel():
    
    def __init__(self, data):
        self.dp = data
        self.pars = self.init_pars()
        
    def init_pars(self):
        fp = FitParameters([('A', 1), ('B', 1), ('omega_a', 1), ('omega_b')])
        return fp

    def __call__(self, p, dp=None, jac=False):
        """
        """
        self.pars.free = p
        
        if dp is None:
            dp = self.dp
        x,y,ey = dp.wavelength, dp.flux, dp.flux_err
        N = len(x)
        
        A = self.pars['A'].full[0]
        B = self.pars['B'].full[0]
        omega_a = self.pars['omega_a'].full[0]
        omega_b = self.pars['omega_b'].full[0]        
        
        v = A * np.sin(omega_a*x) + B * np.cos(omega_b*x)

        if jac:
            ia = np.arange(N)
            ja = np.full(N, 0)
            dfdA = np.sin(omega_a*x)
            
            ib = np.arange(N)
            jb = np.full(N, 1)
            dfdB = np.cos(omega_b*x)

            iom_a = np.arange(N)
            jom_a = np.full(N, 2)
            dfdom_a = A*x*np.cos(omega_a*x)

            iom_b = np.arange(N)
            jom_b = np.full(N, 3)
            dfdom_b = - B*x*np.sin(omega_b*x)
            
            i = np.hstack((ia,ib,iom_a, iom_b))
            j = np.hstack((ja,jb,jom_a, jom_b))
            data = -np.hstack((dfdA, dfdB, dfdom_a, dfdom_b)) / ey[i]
            idx = self.pars.indexof(j)>=0
            
            jacobian = scipy.sparse.coo_matrix((data[idx], (i[idx],j[idx])),
                                        shape=(N,4))
            
            return (y-v)/ey, jacobian

        return (y-v)/ey


class SinCosToyConstraints():
    """
    Constraints on the model parameters
    """
    def __init__(self, pars, vals=[1.2, 0.2]):
        self.pars = pars
        n = len(self.pars.full)
        i = [0, 1, 2, 3]
        j = [0, 0, 1, 1]
        v = [1, -1., 1, -1.]
        self.H = scipy.sparse.coo_matrix((v, (i,j)), shape=(n,2)).tocsr()
        self.d = self.vals = np.array(vals)
        
    #    def __len__(self):
    #        return 2

    def get_rhs(self):
        return self.d

    def __call__(self, p, jac=False):
        self.pars.free = p
        v = -self.H.T @ self.pars.full + self.vals
        if jac:
            i = np.where(self.pars.indexof() >= 0)[0]
            return v, self.H[i,:].T
        return v


    
def main(A=1., B=-0.2, omega_a=0.72, omega_b=0.52,
         minimizer=minimize_nr,#_var_quad,
         plot=False, seed=0):
    """
    """
    dp = gen_sincos(A=A, B=B, omega_a=omega_a, omega_b=omega_b, seed=seed)
    f = SinCosToyModel(dp)
    ff = salt3.models.VarModelResiduals(dp, f)
    cons = SinCosToyConstraints(f.pars)
    x0 = np.array([10., 10., 0.8, 0.8])
    x = minimizer(ff, x0, cons=cons)
    if len(x) == 2:
        x = x[0]
    f.pars.free = x
    print(plot)
    if plot:
        import pylab as pl
        pl.ion()
        import matplotlib as mpl
        pl.figure()
        gs = mpl.gridspec.GridSpec(4,1)
        pl.subplot(gs[0:3,0])
        pl.errorbar(dp.wavelength, dp.flux, yerr=dp.flux_err, ls='', marker='.', color='b')
        xx = np.linspace(dp.wavelength.min(), dp.wavelength.max(), 100)
        xx.sort()
        ddp = gen_sincos(x=xx,
                         y=np.full(len(xx), 0),
                         yerr=np.full(len(xx), 1))
        v = f(x, dp=ddp)
        pl.plot(ddp.wavelength, -v, 'r-')
        pl.ylabel('y')
        
        pl.subplot(gs[3,0])
        pl.errorbar(dp.wavelength, f(x)*dp.flux, yerr=dp.flux_err, marker='.', ls='', color='b')
        pl.grid(1)
        pl.xlabel('x')
        pl.ylabel('res')
        pl.subplots_adjust(hspace=0.05)

    return dp, f, cons, x
    

def test_sincosfit_nr():
    # old target x_target = [0.99669018, -0.20330982,  0.72477005,  0.52477005]
    # first minimizer: minimize_nr : old result was x = [ 0.955678, -0.244322,  0.724236,  0.524236]
    # new minimize_nr_var_quad : x = [0.955678, -0.244322,  0.724236,  0.524236]
    dp, f, cons, x = main(A=1., B=-0.2, omega_a=0.72,
                          omega_b=0.52, minimizer=minimize_nr_var_quad,
                          plot=True, seed=1)
    np.testing.assert_allclose(np.array([0.99669018, -0.20330982,  0.72477005,  0.52477005]), x, atol=1.E-6)
    np.testing.assert_allclose(np.zeros_like(cons(x, jac=False)), cons(x, jac=False), atol=1.E-8)
    
    
def test_sincosfit_nr_qr():
    # another minimizer: minimize_nr_qr [does not scale well with parameter vector size]
    dp, f, cons, x = main(A=1., B=-0.2, omega_a=0.72, omega_b=0.52,
                          minimizer=minimize_nr_var_quad, plot=0, seed=1)
    np.testing.assert_allclose(np.array([ 0.99669018, -0.20330982,  0.72477005,  0.52477005]), x, atol=1.E-6)
    np.testing.assert_allclose(np.zeros_like(cons(x, jac=False)), cons(x, jac=False), atol=1.E-8)
    

if __name__ == "__main__":
    dp, f, cons, x = main(A=1., B=-0.2, omega_a=0.72, omega_b=0.52,
                          minimizer=minimize,#_nr_var_quad,
                          plot=1)



#!/usr/bin/env python3

import os
import sys
from pathlib import Path
import logging
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                    level=logging.INFO)

import pickle
from argparse import ArgumentParser

import matplotlib as mpl
import numpy as np
import numpy.ma as ma
import pylab as pl
import pandas

from lemaitre import bandpasses
from nacl import TrainingDataset, LogLikelihood, Minimizer
from nacl.models.salt2 import SALT2Like, constraints
from nacl.models.salt2.constraints import salt2like_linear_constraints, nacl_linear_constraints
from nacl.models.salt2.regularizations import NaClSplineRegularization
from nacl.models.salt2.variancemodels import SimpleErrorSnake, LocalErrorSnake
from nacl.train import TrainSALT2Like

from saunerie.robuststat import mad




if __name__ == '__main__':

    fl = bandpasses.get_filterlib()
    tds = TrainingDataset.read_parquet('lemaitre.compressed.parquet', filterlib=fl)
    # tds = TrainingDataset.read_parquet('compressed_mini_tds_blinded.parquet', filterlib=fl)

    # this phase range is very small. Find a way to extend it
    # this is mainly because we need the model predictions
    # to calibrate the spectra.
    # clean_dataset(tds, phase_range=(-20., 40.))

    model = SALT2Like(tds) # phase_grid=np.linspace(-30., 70., 30))
    salt2_cons = salt2like_linear_constraints(model)
    nacl_cons = nacl_linear_constraints(model)

    # variance_model = LocalErrorSnake(model)
    #train = TrainSALT2Like(tds, 'sn_lambda_snake')
    #train.train_salt2_model()



r"""
The purpose of regularization is to constrain the surfaces :math:`M_0` and :math:`M_1` in areas of the plane
that are empty of data.
We use a classical regularization scheme, called Generalized Tikhonov Regularization,
consisting in adding a small quadratic penalty.
This penalty is designed to limit the abrupt variations of the model (oscillations)
and to make the model tend slowly towards zero in the absence of data,
in particular at the phases :math:`p < -20` and :math:`p > 60` days.

The :math:`\chi^2_{reg}` is defined as :

.. math::
    \chi^2_{reg} = \mu_{reg} \beta^T P \beta

With :math:`P`, listing parameters in four blocs
([:math:`X_0`, :math:`t_{max}^{B^\star}`, :math:`X_1`, :math:`c`], :math:`M_0`, :math:`M_1`,
[:math:`CL`, :math:`s` , :math:`\eta`, :math:`\kappa`]), by:

.. math::
    \begin{equation}
    P = \left (
    \begin{array}{c c c c}
    0 & 0 & 0 & 0 \\
    0 & P_{\theta_{0|1}} & 0 & 0\\
    0 & 0 & P_{\theta_{0|1}} & 0\\
    0 & 0 & 0 & 0
    \end{array}
    \right)
    \end{equation}

The matrix :math:`P_{\theta_{0|1}}` is defined in the different
regularization class.
"""

import numpy as np
from scipy.sparse import coo_matrix, dia_matrix, csc_matrix, dok_matrix



# this is a minimal rewrite of the original class SplineRegul2D by Guy
# to simplify it (we can condense both regulization inside one single class
# and also to adapt it to the interface expected by the LogLikelihood class.

class SplineRegularization:
    """
    """
    def __init__(self, model, pars, to_regularize=None, mu=1.E-6):
        """
        """
        self.model = model
        # self.pars = self.model.pars
        self.to_regularize = to_regularize
        if to_regularize is None:
            self.to_regularize = ['M0', 'M1']
        self.mu = mu
        # we call it P_full, to remind us that it regularizes all
        # the parameters (free and fixed) of the selected blocks
        self.P_full = self._init_regularization_matrices(pars)

    def _init_regularization_matrices(self, pars):
        raise NotImplementedError('SplineRegularization is pure virtual')

    # def get_free_mat(self, pars):
    #     n = len(self.model.pars.full)
    #     assert n == self.P_full.shape[0]
    #     idx = self.model.pars.indexof(np.arange(n)) >= 0
    #     return self.P_full[:,idx][idx,:]

    def __call__(self, pars, deriv=False):
        """
        """
#        try:
#            self.model.pars.free = p.free
#        except:
#            self.model.pars.free = p.free[:-1]

        full_pars = pars.full
        penality = float(full_pars.T @ self.P_full @ full_pars)
        if not deriv:
            return self.mu * penality

        # gradient of the free parameters
        n = self.P_full.shape[0]
        idx = pars.indexof(np.arange(n)) >= 0
        grad = -2. * (self.P_full @ full_pars)[idx]

        # and finally, the hessian (needs to be sliced also)
        if idx.sum() > 0:
            P = self.P_full[:,idx][idx,:]
        else:
            P = self.P_full
        hess = 2. * P

        mu = self.mu

        return mu * penality, mu * grad, mu * hess


def get_naked_regularization_matrix(size, order=0):
    """
    """
    if order == 0:
        data = np.ones(size)
        return dia_matrix((data, [0]), shape=(size,size)).tocoo()

    data = np.vstack(([3.] * size, [-1.] * size, [-1.] * size))
    data[0, 0] = data[0, -1] = 2.
    mat = dia_matrix((data, [0, -1, 1]), shape=(size,size)).tocoo()
    return mat


class NaClSplineRegularization(SplineRegularization):

    def __init__(self, model, pars, to_regularize, mu=1.E-6, order=1):
        """
        """
        self.order = order
        super().__init__(model, pars, to_regularize, mu)

    def _init_regularization_matrices(self, pars):
        """Build a NaCl regularization matrix
        """
        pars_full = pars.copy()
        pars_full.release()
        npars = len(pars_full.full)
        # if self.p != None:
        #     pars_full = self.p.copy()
        #     pars_full.release()
        #     npars = len(pars_full.full)
        # else :
        #     pars_full = self.model.pars.copy()
        #     pars_full.release()
        #     npars = len(pars_full.full)

        P_full = dok_matrix((npars,npars))

        for block_name in self.to_regularize:
            block_pars = pars_full[block_name].full
            block_size = len(block_pars)
            mat = get_naked_regularization_matrix(block_size, order=self.order)

            # full matrix (used to compute the regularization value)
            i = pars_full[block_name].indexof(mat.row)
            j = pars_full[block_name].indexof(mat.col)
            P_full += coo_matrix((mat.data, (i,j)), shape=(npars,npars))

            # i = pars[block_name].indexof(mat.row)
            # j = pars[block_name].indexof(mat.col)
            # idx = (i >= 0) & (j >= 0)
            # P_free += coo_matrix((mat.data[idx], (i[idx], j[idx])),
            #                      shape=(n_free_pars, n_free_pars))

        return P_full


class SplineRegul2D(object):
    r"""
    This regularization focuses on the model variations in phase.
    The density of spectra allows the basis to be constrained correctly, and the density of points
    on the spectra seems to limit the oscillation phenomena.

    The :math:`\chi^2_{reg}` is defined as :

    .. math::
        \chi^2_{reg} &= \mu_{reg} \left[\sum_{i=0}^{N_{\lambda}-1} \sum_{i=0}^{N_{phase}-1}
        \left(\theta_{0|1}^{j, i+1} - \theta_{0|1}^{j, i}\right)^2 + (\theta_{0|1}^{ji})^2 \right] \\

    This gives that by default :math:`P_{\theta_{0|1}}` is given by:

    .. math::
        P_{\theta_{0|1}} = \begin{pmatrix}
        2 & -1 & 0 &  \cdots & & && 0 \\
        -1 & 3 & -1 & \cdots & &  &&0 \\
        \cdots & & & \cdots & & & &\cdots\\
        \cdots & & -1 & 3 & -1 & & &\cdots\\
        \cdots & & & \cdots & & & &\cdots\\
        0 & && \cdots &  &  -1 & 3 & -1 \\
        0 & & &\cdots &  &  & -1 & 2 \\
        \end{pmatrix}

    else :math:`P_{\theta_{0|1}}` is only the identity matrix.

    Attributes
    ----------
    pars : numpy.array
        Model parameter
    mu : float
        Amplitude of the regularization penalty
    matrix_free : scipy.sparse.coo_matrix
        :math:`P_{\theta_{0|1}}` of only the free parameters.
    matrix_full : scipy.sparse.coo_matrix
        :math:`P_{\theta_{0|1}}` of all the parameters.
    """

    def __init__(self, model, surfaces_reg=['M0'], mu=1.E-6, deriv=True):
        """
        Constructor.
        Creation of the :math:`P` matrix.

        Parameters
        ----------
        model : nacl.salt
            Model to regularize.
        surfaces_reg : list
            List of surfaces to regularized
        mu : float
            Amplitude of the contribution to the :math:`\chi^2`
        deriv : bool
            if True :math:`P_{\theta_{0|1}}` is used.
            Else its a diagonal matrix.
        """
        self.pars = model.pars.copy()
        self.mu = mu

        # build the regularization matrix for the spline parameters
        n = len(self.pars['M0'].full)
        if deriv:
            data = np.vstack(([3.] * n, [-1.] * n, [-1.] * n))
            data[0, 0] = data[0, -1] = 2.
            idx_data = [0, -1, 1]
        else:
            # only diagonal term
            data = np.ones(n).astype(float)
            idx_data = [0]

        matrix = dia_matrix((data, idx_data), shape=(n, n)).tocoo()
        # embed it in a larger matrix, that corresponds to the model parameters
        n_pars = len(self.pars.free)
        i = self.pars['M0'].indexof(matrix.row)
        j = self.pars['M0'].indexof(matrix.col)
        idx = (i >= 0) & (j >= 0)
        matrix_free0 = coo_matrix((matrix.data[idx]*self.mu, (i[idx], j[idx])), shape=(n_pars, n_pars))

        if 'M1' in surfaces_reg:
            n1 = len(self.pars['M1'].full)
            if deriv:
                data1 = np.vstack(([3.] * n1, [-1.] * n1, [-1.] * n1))
                data1[0, 0] = data1[0, -1] = 2.
                idx_data1 = [0, -1, 1]
            else:
                # only diagonal term
                data1 = np.ones(n1).astype(float)
                idx_data1 = [0]

            matrix_1 = dia_matrix((data1, idx_data1), shape=(n1, n1)).tocoo()
            i1 = self.pars['M1'].indexof(matrix_1.row)
            j1 = self.pars['M1'].indexof(matrix_1.col)
            idx1 = (i1 >= 0) & (j1 >= 0)
            matrix_free1 = coo_matrix((matrix_1.data[idx1]*self.mu, (i1[idx1], j1[idx1])), shape=(n_pars, n_pars))
            self.matrix_free = (matrix_free0 + matrix_free1)
        else:
            self.matrix_free = matrix_free0

        pp = self.pars.copy()
        pp.release()
        n_pars = len(pp.full)
        i = pp['M0'].indexof(matrix.row)
        j = pp['M0'].indexof(matrix.col)

        if 'M1' in surfaces_reg:
            i1 = pp['M1'].indexof(matrix_1.row)
            j1 = pp['M1'].indexof(matrix_1.col)
            self.matrix_full = coo_matrix((matrix.data*self.mu, (i, j)),
                                          shape=(n_pars, n_pars))+coo_matrix((matrix_1.data * self.mu, (i1, j1)),
                                                                             shape=(n_pars, n_pars))
        else:
            self.matrix_full = coo_matrix((matrix.data*self.mu, (i, j)), shape=(n_pars, n_pars))  # sans les self.mu

    def __call__(self, beta):
        """
        Calculate the regularization contribution to the final :math:`\chi^2`

        Parameters
        ----------
        beta : array
            Model parameters.

        Returns
        -------
        v : float
            Value of the :math:`\chi^2_{reg}`
        matrix_free : scipy.sparse.coo_matrix
            :math:`P_{\theta_{0|1}}`
        """
        self.pars.free = beta.free
        # there is probably a way to compute this faster
        v = np.dot(self.pars.full.T, self.matrix_full * self.pars.full)
        return v, self.matrix_free


class SplineRegul2DSalt3(object):
    r"""
    This regularization is the one defined in SALT3 (Kenworthy et al. 2021)
    It is focused on the model variations in phase and wavelength.
    The density of spectra allows the basis to be constrained correctly, and the density of points
    on the spectra seems to limit the oscillation phenomena.

    The :math:`\chi^2_{reg}` is defined as :

    .. math::
        \chi^2_{reg} = \mu_{reg} \sum_{j=0}^{2N_{\lambda}} \sum_{i=0}^{2N_{phase}} \left[\frac{\left(
        \left.\frac{\partial M_{0|1}}{\partial p} \right|_{p = p_i,
        \lambda = \lambda_j }\right)^2}{N_{eff}(p_i,\lambda_j)} +
        \frac{\left( \left.\frac{\partial M_{0|1}}{\partial \lambda}
        \right|_{p = p_i, \lambda = \lambda_j } \right)^2}{N_{eff}(p_i,\lambda_j)}
        + \frac{1}{N_{eff}(p_i,\lambda_j)}\left( \frac{\partial M_{0|1}}{\partial p}
        \times \frac{\partial M_{0|1}}{\partial \lambda} + M_{0|1}\frac{\partial^2 M_{0|1}}{\partial
        \lambda \partial p} \right) \right]
    """
    def __init__(self, model, surfaces_reg=['M0'], mu=1.E-6):
        r"""
        Constructor.
        Creation of the :math:`P` matrix.

        Parameters
        ----------
        model : nacl.salt
            Model to regularize.
        surfaces_reg : list
            List of surfaces to regularized
        mu : float
            Amplitude of the contribution to the :math:`\chi^2`
        """
        self.model = model
        self.pars = model.pars.copy()
        self.mu = mu

        self.ph1_range, self.wl1_range, self.N_eff = None, None, None

        self.surfaces_reg = surfaces_reg

        self.wl_range = np.linspace(self.model.wl_range[0], self.model.wl_range[1],
                                    self.model.basis.bx.nj)
        self.ph_range = np.linspace(self.model.phase_range[0], self.model.phase_range[1],
                                    self.model.basis.by.nj)

        self.SP = self.model.training_dataset.spec_data
        # build the regularization matrix for the spline parameters

    def calculate_n_eff(self):
        """
        Calculate the number of data point per bin of wavelength and phase.
        """
        yy_sp, xx_sp = [], []
        bins = self.wl_range
        for sp in self.model.training_dataset.spectra:
            x_sp = sp.data['Wavelength']/(1+sp.data['ZHelio'])
            y_sp = (sp.data['Date']-self.pars['tmax'].free[sp.data['sn_id']])/(1+sp.data['ZHelio'])
            idx_bins = np.where((x_sp.min() < bins) & (x_sp.max() > bins))
            xx_sp.append(bins[idx_bins])
            yy_sp.append(y_sp[0] * np.ones_like(idx_bins[0]))

        xx_sp = np.hstack(xx_sp)
        yy_sp = np.hstack(yy_sp)

        self.ph1_range = np.hstack((self.ph_range, self.ph_range[-1]+1))
        self.wl1_range = np.hstack((self.wl_range, self.wl_range[-1]+1))

        self.N_eff, _, _ = np.histogram2d(xx_sp, yy_sp, bins=[self.wl1_range, self.ph1_range])
        self.N_eff += 1

    def __call__(self, beta):
        """
        Calculate the regularization contribution to the final :math:`\chi^2`

        Parameters
        ----------
        beta : array
            Model parameters.

        Returns
        -------
        v : float
            Value of the :math:`\chi^2_{reg}`
        matrix_free : scipy.sparse.coo_matrix
            :math:`P_{\theta_{0|1}}`
        """
        self.pars.free = beta.free
        self.calculate_n_eff()

        x, y = np.meshgrid(self.wl_range, self.ph_range)
        model_eval = self.model.basis.eval(x.ravel(), y.ravel())
        jacobian_model_eval_wavelength, jacobian_model_eval_phase = self.model.basis.gradient(x.ravel(), y.ravel())

        _, _, hessian_model_eval_phase = self.model.basis.hessian(x.ravel(), y.ravel())

        ddiag = jacobian_model_eval_wavelength * jacobian_model_eval_phase - model_eval * hessian_model_eval_phase

        jacobian_model_eval_phase.data *= jacobian_model_eval_phase.data * self.mu * 0.75/1000  # diff or not
        jacobian_model_eval_wavelength.data *= jacobian_model_eval_wavelength.data * self.mu
        jacobian_model_eval_phase.data /= self.N_eff.ravel()[jacobian_model_eval_phase.row]
        jacobian_model_eval_wavelength.data /= self.N_eff.ravel()[jacobian_model_eval_wavelength.row]

        ddiag.data *= self.mu/self.N_eff.ravel()[ddiag.tocoo().row]  # diff or not

        matrix = (jacobian_model_eval_phase + jacobian_model_eval_wavelength + ddiag)
        matrix = matrix.tocoo()
        n_pars = len(self.pars.free)
        matrix_free = coo_matrix((np.zeros_like(matrix.data), (matrix.row, matrix.col)), shape=(n_pars, n_pars))

        if 'M0' in self.surfaces_reg:
            i = self.pars['M0'].indexof(matrix.row)
            j = self.pars['M0'].indexof(matrix.col)
            idx = (i >= 0) & (j >= 0)
            matrix_free0 = coo_matrix((matrix.data[idx], (i[idx], j[idx])), shape=(n_pars, n_pars))
            matrix_free += matrix_free0

        if 'M1' in self.surfaces_reg:
            # n1 = len(self.pars['M1'].full)
            i1 = self.pars['M1'].indexof(matrix.row)
            j1 = self.pars['M1'].indexof(matrix.col)
            idx1 = (i1 >= 0) & (j1 >= 0)
            matrix_free1 = coo_matrix((matrix.data[idx1], (i1[idx1], j1[idx1])), shape=(n_pars, n_pars))
            matrix_free += matrix_free1

        pp = self.pars.copy()
        pp.release()
        n_pars = len(self.pars.full)
        matrix_full = coo_matrix((np.zeros_like(matrix.data), (matrix.row, matrix.col)), shape=(n_pars, n_pars))

        if 'M0' in self.surfaces_reg:
            i = self.pars['M0'].indexof(matrix.row)
            j = self.pars['M0'].indexof(matrix.col)
            idx = (i >= 0) & (j >= 0)
            matrix_full0 = coo_matrix((matrix.data[idx], (i[idx], j[idx])), shape=(n_pars, n_pars))
            matrix_full += matrix_full0

        if 'M1' in self.surfaces_reg:
            # n1 = len(self.pars['M1'].full)
            i1 = self.pars['M1'].indexof(matrix.row)
            j1 = self.pars['M1'].indexof(matrix.col)
            idx1 = (i1 >= 0) & (j1 >= 0)
            matrix_full1 = coo_matrix((matrix.data[idx1], (i1[idx1], j1[idx1])), shape=(n_pars, n_pars))
            matrix_full += matrix_full1

        self.matrix_full = matrix_full
        self.matrix_free = matrix_free

        v = np.dot(self.pars.full.T, self.matrix_full * self.pars.full)
        return v, self.matrix_free



"""The photometric part of the model

(light curve evaluation)
"""

import time
import numpy as np
import scipy




class LightcurveEvalUnitTz(object):
    """Compute an SN light curve of one supernova in one single band.

    This class is one of the two types of "computing unit". Given a set of
    dates, and the :math:`(X_0, X_1, c, t_{max})` parameters of the SN, compute
    the quantities:

    .. math::

         X_0 \times (1+z) \times \left[I_0 + X_1 I_1\right]

    with

    .. math::
         I_{0|1} = \int M_{0|1}(\lambda, \mathrm{p}) \frac{\lambda}{hc} T((1+z)\lambda) 10^{0.4\ c\ P(\lambda)} d\lambda

    In practice, we can speed up the computations by decomposing the filter on
    a spline basis: :math:`T(\lambda (1+z)) = \sum_q t_q(z) B_q(\lambda)=
    \sum_q (1+z) t_q B_q(\lambda)`, which allows to decompose the integral
    above as:

    .. math::
        I_{0|1} = \sum_{k\ell q} \theta_{0|1, k\ell} t_q(z) B_k(\mathrm{p})\ \int B_\ell(\lambda)
        \frac{\lambda}{hc} B_q(\lambda)\ 10^{0.4\ c\ P(\lambda)} d\lambda

    We use the fact that the color law is a slowly variable function of
    :math:`\lambda` (compared to the individual splines) to expel the color law
    from the integral:

    .. math::
       \int B_\ell(\lambda) \frac{\lambda}{hc} B_q(\lambda) CL(\lambda) d\lambda \approx
       CL(\lambda^{\mathrm{eff}}_{\ell q})\ \times \int B_\ell(\lambda) \frac{\lambda}{hc} B_q(\lambda) d\lambda

    up to second order, if we define:

    .. math::
       \lambda^{\mathrm{eff}}_{\ell q} = \frac{\int B_\ell(\lambda)
       \lambda^2 B_q(\lambda) d\lambda}{\int B_\ell(\lambda) \lambda B_q(\lambda) d\lambda}


    and so, defining: :math:`\Lambda^{\mathrm{eff}} =
    (\lambda^{\mathrm{eff}}_{\ell q})` and ::math:`G = (G_{\ell q}) = (\int
    B_\ell(\lambda)\frac{\lambda}{hc} B_q(\lambda)) d\lambda`, the integral
    evaluation reduces to:

    .. math::
        I_{0|1} = J\ \cdot \Theta\ \cdot \left[G \Lambda^{\mathrm{eff}}\right] \cdot t_z

    where :math:`J`, :math:`G`,
    :math:`\Lambda^{\mathrm{eff}}` and :math:`t_z` are sparse
    and where :math:`G, \Lambda^{\mathrm{eff}}`
    can be precomputed once for all.

    Matrix :math:`G` is called the Gramian of the model and filter spline
    bases. It is precomputed exactly, using Gaussian quadrature when the model
    is instantiated.

    This strategy decompose :math:`T((1+z)\lambda)` and work with a single
    Gramian and an SN-dependant :math:`t_z` vector. This option is probably
    computationally less intensive, as we can factorize the evaluation of the
    color law.

    """

    def __init__(self, lcdata, model):
        """Constructor
        Retrieve the band shape :math:`T(\lambda)` and project :math:`T(\lambda(1+z))` on the basis

        Parameters
        ----------
        lcdata : nacl.dataset.LcData
            Photometric of one Lc of one SN nacl data class.
        model : nacl.models.salt.SALT2Like
            Model.
        """
        self.lcdata = lcdata
        # self.data = lcdata.data # we no longer use this
        self.model = model
        self.z = lcdata.z[0]
        self.band = lcdata.band[0]
        tr = model.filter_db.transmission_db[self.band]
        self.tqz, _ = model.filter_db.insert(tr, z=self.z)
        # set to zero very small values
        self.tqz[(np.abs(self.tqz)/self.tqz.max() < 1e-10)] = 0.

        # self.sn_id = lcdata.sn_id
        self.sn_index = lcdata.sn_index[0]

        #        self.wl_basis_size = len(model.basis.bx)
        self.ph_basis_size = len(model.basis.by)

        self.M0 = model.pars['M0'].full.reshape(self.ph_basis_size, -1)
        self.M1 = model.pars['M1'].full.reshape(self.ph_basis_size, -1)
        self.cl_pars = model.pars['CL'].full

    def __call__(self, jac=False, debug_mode=False):
        """
        Evaluate the SN light curve.

        Parameters
        ----------
        jac : bool
            Whether it return the jacobian matrix
        debug_mode : bool
            if True, return the model components instead of the
            model results.

        Returns
        ----------
        val : numpy.array
            Model evaluation
        jacobian : scipy.sparse.csr_matrix
            Jacobian matrix (optional)

        """
        t0 = time.perf_counter()
        sn_index = self.sn_index
        zz = 1. + self.z
        pars = self.model.pars
        # data = self.lcdata.data
        # band_index = self.lcdata.band_index[0]

        # band_id = self.lcdata.data['band_id'][0]
        # print(self.model.bands[band_id], self.lcdata.data['Filter'][0])
        # lc_id = self.lcdata.data['lc_id'][0]
        # lc_index = self.lcdata.lc_index[0]
        sn_index = self.lcdata.sn_index[0]

        # sn-related parameters
        x0, x1 = pars['X0'].full[sn_index], pars['X1'].full[sn_index]
        c, tmax = pars['col'].full[sn_index],  pars['tmax'].full[sn_index]
        if self.model.calib_error_model is not None:
            # eta = pars['eta_calib'].full[band_index]
            calib_corr = self.model.calib_error_model.correction(lcdata=self.lcdata)
        else:
            calib_corr = np.ones(len(self.lcdata))
        if self.model.color_scatter_model is not None:
            # kappa = pars['kappa_color'].full[lc_index]
            cs_corr = self.model.color_scatter_model.correction()
        else:
            cs_corr = 1.
        # model_to_meas_scale = self.lc_data.norm
        flux_scale = self.model.norm * self.lcdata.norm

        phase_eval = self.model.phase_eval[self.lcdata.slc]
        # color law evaluation -- on the Gram \lambda_eff's
        cl = np.power(10., 0.4 * c * self.model.polynome_color_law.data)
        csr_color_law = scipy.sparse.csr_matrix((cl, (self.model.L_eff.row, self.model.L_eff.col)),
                                                shape=self.model.L_eff.shape)
        gram = self.model.gram.multiply(csr_color_law)
        tqz = self.tqz

        # much faster with the .dot's (instead of @)
        integral_surface_0 = phase_eval.dot(self.M0.dot(gram.dot(tqz)))
        integral_surface_1 = phase_eval.dot(self.M1.dot(gram.dot(tqz)))
        flux = flux_scale * zz * x0 * (integral_surface_0+x1*integral_surface_1) * calib_corr * cs_corr  # (1+eta)*(1+kappa)

        # if debug_mode:
        #     return model_norm * integral_surface_0, model_norm * integral_surface_1, \
        #            cl, flux, x0, x1, model_norm

        self.model.val.append(flux)
        if not jac:
            self.model.timing.append(time.perf_counter()-t0)
            return flux

        # n_data = len(self.lcdata)
        # sn_index = np.full(n_data, self.sn_index)
        # norm = self.model.norm

        # shortcut names to the internal cache holding the
        # jacobian matrix definition
        jacobian_i, jacobian_j, jacobian_val = self.model.jacobian_i, self.model.jacobian_j, self.model.jacobian_val

        # dMdX0
        jacobian_i.append(self.lcdata.row)
        jacobian_j.append(self.model.pars['X0'].indexof(self.lcdata.sn_index))
        assert len(self.lcdata.row) == len(self.model.pars['X0'].indexof(self.lcdata.sn_index))
        jacobian_val.append(flux_scale * zz * (integral_surface_0 + x1*integral_surface_1) * calib_corr * cs_corr)  # (1+eta) * (1+kappa))

        # dMdX1
        jacobian_i.append(self.lcdata.row)
        jacobian_j.append(self.model.pars['X1'].indexof(self.lcdata.sn_index))
        assert len(self.lcdata.row) == len(self.model.pars['X1'].indexof(self.lcdata.sn_index))
        jacobian_val.append(flux_scale * zz * x0 * integral_surface_1 * calib_corr * cs_corr)  # (1+eta) * (1+kappa))

        # dMdc
        jacobian_i.append(self.lcdata.row)
        jacobian_j.append(self.model.pars['col'].indexof(self.lcdata.sn_index))
        assert len(self.lcdata.row) == len(self.model.pars['col'].indexof(self.lcdata.sn_index))
        gram_color_law = gram.multiply(self.model.polynome_color_law)
        dintegral0_dc = 0.4 * np.log(10.) * phase_eval.dot(self.M0.dot(gram_color_law.dot(tqz)))
        dintegral1_dc = 0.4 * np.log(10.) * phase_eval.dot(self.M1.dot(gram_color_law.dot(tqz)))
        jacobian_val.append(flux_scale * zz * x0 * (dintegral0_dc + x1*dintegral1_dc) * calib_corr * cs_corr)  # (1+eta) * (1+kappa))

        # dMdtmax
        dphase_dtmax = self.model.dphase_dtmax[self.lcdata.slc]
        jacobian_i.append(self.lcdata.row)
        jacobian_j.append(self.model.pars['tmax'].indexof(self.lcdata.sn_index))
        assert len(self.lcdata.row) == len(self.model.pars['tmax'].indexof(self.lcdata.sn_index))
        dintegral0_dtmax = -dphase_dtmax.dot(self.M0.dot(gram.dot(tqz))) / zz
        dintegral1_dtmax = -dphase_dtmax.dot(self.M1.dot(gram.dot(tqz))) / zz
        jacobian_val.append(flux_scale * zz * x0 * (dintegral0_dtmax + x1*dintegral1_dtmax) * calib_corr * cs_corr)  # (1+eta)*(1+kappa))

        # dMdtheta_0
        dbase_dtheta = scipy.sparse.kron(phase_eval, gram.dot(tqz)).tocoo()
        jacobian_i.append(self.lcdata.row[dbase_dtheta.row])
        jacobian_j.append(self.model.pars['M0'].indexof(dbase_dtheta.col))
        assert len(self.lcdata.row[dbase_dtheta.row]) == len(self.model.pars['M0'].indexof(dbase_dtheta.col))
        jacobian_val.append(flux_scale[dbase_dtheta.row] * zz * x0 * dbase_dtheta.data * calib_corr[dbase_dtheta.row] * cs_corr)  # (1+eta) * (1+kappa))

        # dMdtheta_1
        jacobian_i.append(self.lcdata.row[dbase_dtheta.row])
        jacobian_j.append(self.model.pars['M1'].indexof(dbase_dtheta.col))
        assert len(self.lcdata.row[dbase_dtheta.row]) == len(self.model.pars['M1'].indexof(dbase_dtheta.col))
        jacobian_val.append(flux_scale[dbase_dtheta.row] * zz * x0 * x1 * dbase_dtheta.data * calib_corr[dbase_dtheta.row] * cs_corr)  # (1+eta) * (1+kappa))

        # dMdcl
        # I really don't know how to vectorize the computation
        # of these derivatives. If somebody has an idea let me know.
        buff = np.zeros(len(self.lcdata)).astype(int)
        for i, jacobian_color_law in enumerate(self.model.jacobian_color_law):
            buff[:] = i
            jacobian_i.append(self.lcdata.row)
            jacobian_j.append(self.model.pars['CL'].indexof(buff))
            gram_jacobian_color_law = gram.multiply(jacobian_color_law)
            jacobian_surface0_color_law = 0.4 * np.log(10.) * c * phase_eval.dot(self.M0.dot(
                gram_jacobian_color_law.dot(tqz)))
            jacobian_surface1_color_law = 0.4 * np.log(10.) * c * phase_eval.dot(self.M1.dot(
                gram_jacobian_color_law.dot(tqz)))
            jacobian_val.append(flux_scale * zz * x0 * (jacobian_surface0_color_law + x1*jacobian_surface1_color_law)
                                * calib_corr * cs_corr)  # * (1+eta) * (1+kappa))

        # dMdeta
        if self.model.calib_error_model is not None:
            jacobian_i.append(self.lcdata.row)
            jacobian_j.append(self.model.pars['eta_calib'].indexof(self.lcdata.band_index))
            jacobian_val.append(flux_scale * x0 * zz * (integral_surface_0 + x1*integral_surface_1) * cs_corr)  # *(1+kappa))

        # dKappa
        if self.model.color_scatter_model is not None:
            jacobian_i.append(self.lcdata.row)
            jacobian_j.append(self.model.pars['kappa_color'].indexof(self.lcdata.band_index))
            jacobian_val.append(flux_scale * x0 * zz * (integral_surface_0 + x1*integral_surface_1) * calib_corr)  # (1+eta))

        if self.model.disable_cache:
            self.model.clear_cache()
        self.model.timing.append(time.perf_counter()-t0)
        return flux



"""
"""

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
import pandas
from ....lib.dataproxy import DataProxy
from ....lib.bspline import BSpline
from .model import Model




def gaussian_2d(x, y, sigma):
    """
    Gaussian function *2
    """
    #norm = a / (np.sqrt(2. * np.pi) * sigma)
    xx = x / sigma
    yy = y / sigma
    return np.exp(-0.5 * (xx**2 + yy**2))


class ToyModelGenerator:

    def __init__(self, nsn=100, npts1=10, npts2=10, nbands=1,
                 x1_min=-10., x1_max=10., x2_min=-10., x2_max=10., delta=10.,
                 yerr=0.01, error_pedestal=None, error_snake=None, wl_pivot=5000.):
        """
        .. note::
           error snake and error_pedestal are mutually exclusive.
           if error_snake > 0, then error_pedestal is set to zero
        """
        self.nsn = nsn
        self.npts1 = npts1
        self.npts2 = npts2
        self.nbands = nbands
        self.x1_min = x1_min
        self.x1_max = x1_max
        self.x2_min = x2_min
        self.x2_max = x2_max
        self.delta = delta
        self.yerr = yerr
        self.error_pedestal = 0.
        self.error_snake = 0.
        if error_pedestal is not None:
            self.error_pedestal = error_pedestal
        if error_snake is not None:
            self.error_pedestal = 0.
            self.error_snake = error_snake
        self.wl_pivot = wl_pivot
            
        self.sn_pars = {'x0_range': (0.1, 100.), 'tmax_margin': 5., 'stretch_range': (0., 0.25), 'color_pars': (0., 0.1)}
        self.lcpars = {'g': (0., 0.8), 'r': (0.2, 1.2), 'i': (0.7, 2.)}

    def gen_sample(self):
        """
        """
        # values and errors
        if 'x0_range' in self.sn_pars:
            xmin, xmax = self.sn_pars['x0_range']
            x0 = np.random.uniform(xmin, xmax, size=self.nsn)
        else:
            x0 = np.full(self.nsn, 1.)
        if 'tmax_margin' in self.sn_pars:
            tmax = np.random.uniform(self.x1_min+5, self.x1_max-5, size=self.nsn)
        else:
            tmax = np.full(self.nsn, 0.)
        if 'stretch_range' in self.sn_pars:
            stretch = np.random.normal(0., 0.25, size=self.nsn)
        else:
            stretch = np.full(self.nsn, 0.)

        if 'color_pars' in self.sn_pars:
            loc, sig = self.sn_pars['color_pars']
            color = np.random.normal(loc=loc, scale=sig, size=self.nsn)

        sample = pandas.DataFrame({'x0': x0, 'tmax': tmax, 'stretch': stretch, 'color': color})
        return sample

    def gen_data(self, sample):
        """
        """
        # generate the data structure
        npts = self.nsn * self.nbands * self.npts1 * self.npts2
        zz = np.random.uniform(1.0, 1.1, self.nsn)
        d = {
            'x1': np.random.uniform(self.x1_min, self.x1_max, npts),
            'x2': np.random.uniform(self.x2_min, self.x2_max, npts),
            'sn': np.repeat(np.arange(self.nsn), self.nbands * self.npts1 * self.npts2),
            'band': np.tile(np.repeat(np.arange(self.nbands), self.npts1 * self.npts2), self.nsn),
            'y': np.zeros(npts),
            'yerr': np.zeros(npts),
            'zz': np.repeat(zz, self.nbands * self.npts1 * self.npts2)
        }

        for band in self.lcpars:
            d['wl'] = 1500. * d['band'] + 4000.
            dwl = (d['wl'] - self.wl_pivot) / self.wl_pivot
            cl = 1. + dwl * sample.color[d['sn']]
            dt, sigma = self.lcpars[band]
            ph1 = (d['x1']-sample.tmax[d['sn']]-dt) * (1. + sample.stretch[d['sn']]) / d['zz'][d['sn']]
            ph2 = (d['x2']-sample.tmax[d['sn']]-dt) * (1. + sample.stretch[d['sn']]) / d['zz'][d['sn']]
            d['y'] = sample.x0[d['sn']] * gaussian_2d(ph1, ph2, sigma) / d['zz'][d['sn']]
            d['yerr'] = sample.x0[d['sn']] * np.full(npts, self.yerr)
            d['var'] = np.sqrt(self.error_pedestal**2 + self.error_snake**2 * d['y']**2)
        d = pandas.DataFrame(d)

        # add noise
        sig = np.sqrt(d.yerr**2 + self.error_pedestal**2 + self.error_snake**2 * d['y']**2)
        d['noise'] = np.random.normal(scale=sig, size=len(sig))
        d.y += d['noise']

        dp = DataProxy(d, x1='x1', x2='x2', sn='sn', band='band', wl='wl', y='y', yerr='yerr', var='var', zz='zz')
        dp.add_field('bads', np.zeros(len(dp.nt)).astype(bool))
        dp.make_index('sn')
        dp.make_index('band')

        return dp
    
    def gen_fine_grid(self, npts_per_lc=100):
        """
        """
        npts = 1 * self.nbands * npts_per_lc
        d = {
            'x1': np.tile(np.linspace(self.x1_min, self.x1_max, npts_per_lc), self.nbands),
            'x2': np.tile(np.linspace(self.x2_min, self.x2_max, npts_per_lc), self.nbands),
            'sn': np.zeros(npts).astype(int),
            'band': np.repeat(np.arange(self.nbands), npts_per_lc).astype(int),
            'y': np.zeros(npts),
            'yerr': np.zeros(npts),
            'var': np.zeros(npts),
            'bads': np.zeros(npts).astype(bool)
        }
        d = pandas.DataFrame(d)
        d['wl'] = 1500. * d['band'] + 4000.

        dp = DataProxy(d, x1='x1', x2='x2', sn='sn', band='band', wl='wl', y='y', yerr='yerr', var='var')
        dp.make_index('sn')
        dp.make_index('band')

        return dp
    
    def generate(self):
        """
        """
        sample = self.gen_sample()
        dp = self.gen_data(sample)
        dp.sample = sample
        return dp
